#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-reschedule
# Author :  Stuart Paterson
########################################################################
"""
Reschedule the given DIRAC job

Example:
  $ dirac-wms-job-reschedule 1
  Rescheduled job 1
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    result = dirac.rescheduleJob(parseArguments(args))
    if result["OK"]:
        print("Rescheduled job %s" % ",".join([str(j) for j in result["Value"]]))
    else:
        errorList.append((result["Value"][-1], result["Message"]))
        print(result["Message"])
        exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Script to run Executable application"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import subprocess
import shlex

# Main
if __name__ == "__main__":

    sys.exit(subprocess.call(shlex.split("echo Hello World")))



"""Test the API class."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

from DIRAC.Core.Base.API import API


def test_deepcopy():
    """Ensure we can copy the API classes."""
    instance_A = API()
    instance_B = copy.deepcopy(instance_A)
    assert instance_B.log



########################################################################
# File :    InputDataByProtocol.py
# Author :  Stuart Paterson
########################################################################

""" The Input Data By Protocol module wraps around the Replica Management
    components to provide access to datasets by available site protocols as
    defined in the CS for the VO.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient

COMPONENT_NAME = "InputDataByProtocol"


class InputDataByProtocol(object):

    #############################################################################
    def __init__(self, argumentsDict):
        """Standard constructor"""
        self.name = COMPONENT_NAME
        self.log = gLogger.getSubLogger(self.name)
        self.inputData = argumentsDict["InputData"]
        self.configuration = argumentsDict["Configuration"]
        self.fileCatalogResult = argumentsDict["FileCatalog"]
        self.jobID = None
        # This is because  replicas contain SEs and metadata keys!
        # FIXME: the structure of the dictionary must be fixed to avoid this mess
        self.metaKeys = set(
            [
                "ChecksumType",
                "Checksum",
                "NumberOfLinks",
                "Mode",
                "GUID",
                "Status",
                "ModificationDate",
                "CreationDate",
                "Size",
                "Owner",
                "OwnerGroup",
                "GID",
                "UID",
                "FileID",
            ]
        )

    #############################################################################
    def execute(self, dataToResolve=None):
        """This method is called to obtain the TURLs for all requested input data
        firstly by available site protocols and redundantly via TURL construction.
        If TURLs are missing these are conveyed in the result to
        """

        # Define local configuration options present at every site
        localSEList = self.configuration["LocalSEList"]
        self.jobID = self.configuration.get("JobID")
        allReplicas = self.configuration.get("AllReplicas", False)
        if allReplicas:
            self.log.info("All replicas will be used in the resolution")

        if dataToResolve:
            self.log.verbose("Data to resolve passed directly to InputDataByProtocol module")
            self.inputData = dataToResolve  # e.g. list supplied by another module

        if isinstance(self.inputData, six.string_types):
            self.inputData = self.inputData.replace(" ", "").split(",")

        self.inputData = [x.replace("LFN:", "") for x in self.inputData]
        self.log.verbose("InputData requirement to be resolved by protocol is:\n%s" % "\n".join(self.inputData))

        # First make a check in case replicas have been removed or are not accessible
        # from the local site (remove these from consideration for local protocols)
        replicas = self.fileCatalogResult["Value"]["Successful"]
        self.log.debug("File Catalogue result is:\n%s" % str(replicas))

        # First get the preferred replica:
        requestedProtocol = self.configuration.get("Protocol", "")
        result = self.__resolveReplicas(localSEList, replicas, requestedProtocol=requestedProtocol)
        if not result["OK"]:
            return result
        success = result["Value"]["Successful"]
        if not allReplicas:
            bestReplica = {}
            for lfn in success:
                bestReplica[lfn] = success[lfn][0]
            return S_OK({"Successful": bestReplica, "Failed": result["Value"]["Failed"]})

        # Keep the failed LFNs from local SEs
        failed = set(result["Value"]["Failed"])
        # If all replicas are requested, get results for other SEs
        seList = set()
        localSESet = set(localSEList)
        for lfn in list(replicas):
            extraSEs = set(replicas[lfn]) - localSESet
            # If any extra SE, add it to the set, othewise don't consider that file
            if extraSEs:
                seList.update(extraSEs)
            else:
                replicas.pop(lfn)
        seList -= self.metaKeys

        if seList:
            requestedProtocol = self.configuration.get("RemoteProtocol", "")
            result = self.__resolveReplicas(seList, replicas, ignoreTape=True, requestedProtocol=requestedProtocol)
            if not result["OK"]:
                return result
            for lfn in result["Value"]["Successful"]:
                success.setdefault(lfn, []).extend(result["Value"]["Successful"][lfn])
            failed.update(result["Value"]["Failed"])
        # Only consider failed the files that are not successful as well
        failed -= set(success)
        return S_OK({"Successful": success, "Failed": sorted(failed)})

    def __resolveReplicas(self, seList, replicas, ignoreTape=False, requestedProtocol=""):
        diskSEs = set()
        tapeSEs = set()
        if not seList:
            return S_OK({"Successful": {}, "Failed": []})

        for localSE in seList:
            seStatus = StorageElement(localSE).status()
            if seStatus["Read"] and seStatus["DiskSE"]:
                diskSEs.add(localSE)
            elif seStatus["Read"] and seStatus["TapeSE"]:
                tapeSEs.add(localSE)

        # For the case that a file is found on two SEs at the same site
        # disk-based replicas are favoured.
        # Problematic files will be returned and can be handled by another module
        failedReplicas = set()
        newReplicasDict = {}
        for lfn, reps in replicas.items():
            if lfn in self.inputData:
                # Check that all replicas are on a valid local SE
                if not [se for se in reps if se in diskSEs.union(tapeSEs)]:
                    failedReplicas.add(lfn)
                else:
                    sreps = set(reps)
                    for seName in diskSEs & sreps:
                        newReplicasDict.setdefault(lfn, []).append(seName)
                    if not newReplicasDict.get(lfn) and not ignoreTape:
                        for seName in tapeSEs & sreps:
                            newReplicasDict.setdefault(lfn, []).append(seName)

        # Check that all LFNs have at least one replica and GUID
        if failedReplicas:
            # in principle this is not a failure but depends on the policy of the VO
            # datasets could be downloaded from another site
            self.log.info(
                "The following file(s) were found not to have replicas on any of %s:\n%s"
                % (str(seList), "\n".join(sorted(failedReplicas)))
            )

        # Need to group files by SE in order to stage optimally
        # we know from above that all remaining files have a replica
        # (preferring disk if >1) in the local storage.
        # IMPORTANT, only add replicas for input data that is requested
        # since this module could have been executed after another.
        seFilesDict = {}
        for lfn, seList in newReplicasDict.items():
            for seName in seList:
                seFilesDict.setdefault(seName, []).append(lfn)

        sortedSEs = sorted(((len(lfns), seName) for seName, lfns in seFilesDict.items()), reverse=True)

        trackLFNs = {}
        for _len, seName in sortedSEs:
            for lfn in seFilesDict[seName]:
                if "Size" in replicas[lfn] and "GUID" in replicas[lfn]:
                    trackLFNs.setdefault(lfn, []).append(
                        {
                            "pfn": replicas.get(lfn, {}).get(seName, lfn),
                            "se": seName,
                            "size": replicas[lfn]["Size"],
                            "guid": replicas[lfn]["GUID"],
                        }
                    )

        self.log.debug("Files grouped by SEs are:\n%s" % str(seFilesDict))
        for seName, lfns in seFilesDict.items():
            self.log.info(" %s LFNs found from catalog at SE %s" % (len(lfns), seName))
            self.log.verbose("\n".join(lfns))

        # Can now start to obtain TURLs for files grouped by localSE
        # for requested input data
        for seName, lfns in seFilesDict.items():
            if not lfns:
                continue
            failedReps = set()
            result = StorageElement(seName).getFileMetadata(lfns)
            if not result["OK"]:
                self.log.error("Error getting metadata.", result["Message"] + ":\n%s" % "\n".join(lfns))
                # If we can not get MetaData, most likely there is a problem with the SE
                # declare the replicas failed and continue
                failedReps.update(lfns)
                continue
            failed = result["Value"]["Failed"]
            if failed:
                # If MetaData can not be retrieved for some PFNs
                # declared them failed and go on
                for lfn in failed:
                    lfns.remove(lfn)
                    if isinstance(failed, dict):
                        self.log.error(failed[lfn], lfn)
                    failedReps.add(lfn)
            for lfn, metadata in result["Value"]["Successful"].items():
                if metadata.get("Lost", False):
                    error = "File has been Lost by the StorageElement %s" % seName
                elif metadata.get("Unavailable", False):
                    error = "File is declared Unavailable by the StorageElement %s" % seName
                elif seName in tapeSEs and not metadata.get("Cached", metadata["Accessible"]):
                    error = "File is not online in StorageElement %s Cache" % seName
                elif not metadata.get("Accessible", True):
                    error = "File is not accessible"
                else:
                    error = ""
                if error:
                    lfns.remove(lfn)
                    self.log.error(error, lfn)
                    # If PFN is not available
                    # declared it failed and go on
                    failedReps.add(lfn)

            if None in failedReps:
                failedReps.remove(None)
            if not failedReps:
                self.log.info("Preliminary checks OK, getting TURLS at %s for:\n%s" % (seName, "\n".join(lfns)))
            else:
                self.log.warn("Errors during preliminary checks for %d files" % len(failedReps))

            result = StorageElement(seName).getURL(lfns, protocol=requestedProtocol)
            if not result["OK"]:
                self.log.error("Error getting TURLs", result["Message"])
                return result

            badTURLCount = 0
            badTURLs = []
            seResult = result["Value"]

            for lfn, cause in seResult["Failed"].items():
                badTURLCount += 1
                badTURLs.append("Failed to obtain TURL for %s: %s" % (lfn, cause))
                failedReps.add(lfn)

            if badTURLCount:
                self.log.warn("Found %s problematic TURL(s) for job %s" % (badTURLCount, self.jobID))
                param = "\n".join(badTURLs)
                self.log.info(param)
                result = self.__setJobParam("ProblematicTURLs", param)
                if not result["OK"]:
                    self.log.warn("Error setting job param", result["Message"])

            failedReplicas.update(failedReps)
            for lfn, turl in seResult["Successful"].items():
                for track in trackLFNs[lfn]:
                    if track["se"] == seName:
                        track["turl"] = turl
                        break
                self.log.info("Resolved input data\n>>>> SE: %s\n>>>>LFN: %s\n>>>>TURL: %s" % (seName, lfn, turl))
            ##### End of loop on SE #######

        # Check if the files were actually resolved (i.e. have a TURL)
        # If so, remove them from failed list
        for lfn, mdataList in list(trackLFNs.items()):  # There is a pop below, can't iterate trackLFNs
            for mdata in list(mdataList):
                if "turl" not in mdata:
                    mdataList.remove(mdata)
                    self.log.info("No TURL resolved for %s at %s" % (lfn, mdata["se"]))
            if not mdataList:
                trackLFNs.pop(lfn, None)
                failedReplicas.add(lfn)
            elif lfn in failedReplicas:
                failedReplicas.remove(lfn)
        self.log.debug("All resolved data", sorted(trackLFNs))
        self.log.debug("All failed data", sorted(failedReplicas))

        return S_OK({"Successful": trackLFNs, "Failed": sorted(failedReplicas)})

    #############################################################################
    def __setJobParam(self, name, value):
        """Wraps around setJobParameter of state update client"""
        if not self.jobID:
            return S_ERROR("JobID not defined")

        self.log.verbose("setJobParameter(%s, %s, %s)" % (self.jobID, name, value))
        return JobStateUpdateClient().setJobParameter(int(self.jobID), str(name), str(value))



"""
It is used to load classes from a specific system.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.ObjectLoader import loadObjects



""" SLURM.py is a DIRAC independent class representing SLURM batch system.
    SLURM objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import re
import subprocess
import shlex
import random

__RCSID__ = "$Id$"


class SLURM(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the OAR batch system"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "ErrorDir", "Queue", "SubmitOptions"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs")
        if not nJobs:
            nJobs = 1

        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]
        queue = kwargs["Queue"]
        submitOptions = kwargs["SubmitOptions"]
        executable = kwargs["Executable"]
        numberOfProcessors = kwargs.get("NumberOfProcessors", 1)
        # numberOfNodes is treated as a string as it can contain values such as "2-4"
        # where 2 would represent the minimum number of nodes to allocate, and 4 the maximum
        numberOfNodes = kwargs.get("NumberOfNodes", "1")
        numberOfGPUs = kwargs.get("NumberOfGPUs")
        preamble = kwargs.get("Preamble")

        outFile = os.path.join(outputDir, "%jobid%")
        errFile = os.path.join(errorDir, "%jobid%")
        outFile = os.path.expandvars(outFile)
        errFile = os.path.expandvars(errFile)
        executable = os.path.expandvars(executable)

        # There are more than 1 node, we have to run the executable in parallel on different nodes using srun
        if numberOfNodes != "1":
            executable = self._generateSrunWrapper(executable)

        jobIDs = []
        for _i in range(nJobs):
            jid = ""
            cmd = "%s; " % preamble if preamble else ""
            # By default, all the environment variables of the submitter node are propagated to the workers
            # It can create conflicts during the installation of the pilots
            # --export restricts the propagation to the PATH variable to get a clean environment in the workers
            cmd += "sbatch --export=PATH "
            cmd += "-o %s/%%j.out " % outputDir
            cmd += "-e %s/%%j.err " % errorDir
            cmd += "--partition=%s " % queue
            # One pilot (task) per node, allocating a certain number of processors
            cmd += "--ntasks-per-node=1 "
            cmd += "--nodes=%s " % numberOfNodes
            cmd += "--cpus-per-task=%d " % numberOfProcessors
            if numberOfGPUs:
                cmd += "--gpus-per-task=%d " % int(numberOfGPUs)
            # Additional options
            cmd += "%s %s" % (submitOptions, executable)
            sp = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0 or not output:
                break

            lines = output.split("\n")
            for line in lines:
                result = re.search(r"Submitted batch job (\d*)", line)
                if result:
                    jid = result.groups()[0]
                    break

            if not jid:
                break

            jid = jid.strip()
            jobIDs.append(jid)

        # Delete the srun wrapper
        if numberOfNodes != "1":
            os.remove(executable)

        if jobIDs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = jobIDs
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error
        return resultDict

    def _generateSrunWrapper(self, executableFile):
        """
        Associate the executable with srun, to execute the same command in parallel on multiple nodes.
        Wrap it in a new executable file

        :param str executableFile: name of the executable file to wrap
        :return str: name of the wrapper that runs the executable via srun
        """
        suffix = random.randrange(1, 99999)
        wrapper = "srunExec_%s.sh" % suffix
        with open(executableFile, "r") as f:
            content = f.read()

        # Build the script to run the executable in parallel multiple times
        # - Embed the content of executableFile inside the parallel library wrapper script
        # - srun is the command to execute a task multiple time in parallel
        #   -l option: add the task ID to the output
        #   -k option: do not kill the slurm job if one of the nodes is broken
        cmd = """#!/bin/bash
cat > %(wrapper)s << EOFEXEC
%(content)s
EOFEXEC
chmod 755 %(wrapper)s
srun -l -k %(wrapper)s
""" % dict(
            wrapper=wrapper, content=content
        )
        return cmd

    def killJob(self, **kwargs):
        """Delete a job from SLURM batch scheduler. Input: list of jobs output: int"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList", "Queue"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        queue = kwargs["Queue"]

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            cmd = "scancel --partition=%s %s" % (queue, job)
            sp = subprocess.Popen(
                shlex.split(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            _, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = errors
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get status of the jobs in the given list"""

        resultDict = {}

        if "JobIDList" not in kwargs or not kwargs["JobIDList"]:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        jobIDList = kwargs["JobIDList"]

        jobIDs = ""
        for jobID in jobIDList:
            jobIDs += jobID + ","

        # displays accounting data for all jobs in the Slurm job accounting log or Slurm database
        # -j is the given job
        # -o the information of interest
        # -X to get rid of intermediate steps
        # -n to remove the header
        # -P to make the output parseable (remove tabs, spaces, columns)
        # --delimiter to specify character that splits the fields
        cmd = "sacct -j %s -o JobID,STATE -X -n -P --delimiter=," % jobIDs
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode
        if status != 0:
            resultDict["Status"] = 1
            resultDict["Message"] = error
            return resultDict

        statusDict = {}
        lines = output.strip().split("\n")
        jids = set()
        for line in lines:
            jid, status = line.split(",")
            jids.add(jid)
            if jid in jobIDList:
                if status in ["PENDING", "SUSPENDED", "CONFIGURING"]:
                    statusDict[jid] = "Waiting"
                elif status in ["RUNNING", "COMPLETING"]:
                    statusDict[jid] = "Running"
                elif status in ["CANCELLED", "PREEMPTED"]:
                    statusDict[jid] = "Aborted"
                elif status in ["COMPLETED"]:
                    statusDict[jid] = "Done"
                elif status in ["FAILED", "TIMEOUT", "NODE_FAIL"]:
                    statusDict[jid] = "Failed"
                else:
                    statusDict[jid] = "Unknown"

        leftJobs = set(jobIDList) - jids
        for jid in leftJobs:
            statusDict[jid] = "Unknown"

        # Final output
        resultDict["Status"] = 0
        resultDict["Jobs"] = statusDict
        return resultDict

    def getCEStatus(self, **kwargs):
        """Get the overall status of the CE"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Queue"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        queue = kwargs["Queue"]

        cmd = "squeue --partition=%s --user=%s --format='%%j %%T' " % (queue, user)
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode
        if status != 0:
            resultDict["Status"] = 1
            resultDict["Message"] = error
            return resultDict

        waitingJobs = 0
        runningJobs = 0
        lines = output.split("\n")
        for line in lines[1:]:
            _jid, status = line.split()
            if status in ["PENDING", "SUSPENDED", "CONFIGURING"]:
                waitingJobs += 1
            elif status in ["RUNNING", "COMPLETING"]:
                runningJobs += 1

        # Final output
        resultDict["Status"] = 0
        resultDict["Waiting"] = waitingJobs
        resultDict["Running"] = runningJobs
        return resultDict

    def getJobOutputFiles(self, **kwargs):
        """Get output file names and templates for the specific CE

        Reorder the content of the output files according to the node identifier
        if multiple nodes were involved.

        From:
        >>> 1: line1
        >>> 2: line1
        >>> 1: line2
        To:
        >>> # On node 1
        >>>   line1
        >>>   line2
        >>> # On node 2
        >>>   line1
        """
        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList", "OutputDir", "ErrorDir"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]
        jobIDList = kwargs["JobIDList"]
        numberOfNodes = kwargs.get("NumberOfNodes", "1")

        jobDict = {}
        for jobID in jobIDList:
            output = "%s/%s.out" % (outputDir, jobID)
            error = "%s/%s.err" % (errorDir, jobID)

            if numberOfNodes != "1":
                self._openFileAndSortOutput(output)
                self._openFileAndSortOutput(error)

            jobDict[jobID] = {}
            jobDict[jobID]["Output"] = output
            jobDict[jobID]["Error"] = error

        resultDict["Status"] = 0
        resultDict["Jobs"] = jobDict
        return resultDict

    def _openFileAndSortOutput(self, outputFile):
        """
        Open a file, get its content and reorder it according to the node identifiers

        :param str outputFile: name of the file to sort
        """
        with open(outputFile, "r") as f:
            outputContent = f.read()

        sortedContent = self._sortOutput(outputContent)

        with open(outputFile, "w") as f:
            f.write(sortedContent)

    def _sortOutput(self, outputContent):
        """
        Reorder the content of the output file according to the node identifiers

        :param str outputContent: content to sort
        :return str: content sorted
        """
        outputLines = outputContent.split("\n")
        nodes = {}
        for line in outputLines:
            node, line_content = line.split(":", 1)
            if node not in nodes:
                nodes[node] = []
            nodes[node].append(line_content)

        content = ""
        for node, lines in nodes.items():
            content += "# On node %s\n\n" % node
            content += "\n".join(lines) + "\n"
        return content



########################################################################
# File :   SSHComputingElement.py
# Author : Dumitru Laurentiu
########################################################################

""" SSH (Virtual) Computing Element: For a given list of ip/cores pair it will send jobs
    directly through ssh
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import socket
import stat
from six.moves.urllib.parse import urlparse

from DIRAC import S_OK, S_ERROR
from DIRAC import rootPath

from DIRAC.Resources.Computing.SSHComputingElement import SSHComputingElement
from DIRAC.Resources.Computing.PilotBundle import bundleProxy, writeScript
from DIRAC.WorkloadManagementSystem.Client import PilotStatus


class SSHBatchComputingElement(SSHComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(SSHBatchComputingElement, self).__init__(ceUniqueID)

        self.ceType = "SSHBatch"
        self.sshHost = []
        self.execution = "SSHBATCH"

    def _reset(self):

        batchSystemName = self.ceParameters.get("BatchSystem", "Host")
        if "BatchSystem" not in self.ceParameters:
            self.ceParameters["BatchSystem"] = batchSystemName
        result = self.loadBatchSystem(batchSystemName)
        if not result["OK"]:
            self.log.error("Failed to load the batch system plugin", batchSystemName)
            return result

        self.user = self.ceParameters["SSHUser"]
        self.queue = self.ceParameters["Queue"]
        self.sshScript = os.path.join(rootPath, "DIRAC", "Resources", "Computing", "remote_scripts", "sshce")
        if "ExecQueue" not in self.ceParameters or not self.ceParameters["ExecQueue"]:
            self.ceParameters["ExecQueue"] = self.ceParameters.get("Queue", "")
        self.execQueue = self.ceParameters["ExecQueue"]
        self.log.info("Using queue: ", self.queue)
        self.hostname = socket.gethostname()
        self.sharedArea = self.ceParameters["SharedArea"]
        self.batchOutput = self.ceParameters["BatchOutput"]
        if not self.batchOutput.startswith("/"):
            self.batchOutput = os.path.join(self.sharedArea, self.batchOutput)
        self.batchError = self.ceParameters["BatchError"]
        if not self.batchError.startswith("/"):
            self.batchError = os.path.join(self.sharedArea, self.batchError)
        self.infoArea = self.ceParameters["InfoArea"]
        if not self.infoArea.startswith("/"):
            self.infoArea = os.path.join(self.sharedArea, self.infoArea)
        self.executableArea = self.ceParameters["ExecutableArea"]
        if not self.executableArea.startswith("/"):
            self.executableArea = os.path.join(self.sharedArea, self.executableArea)
        self.workArea = self.ceParameters["WorkArea"]
        if not self.workArea.startswith("/"):
            self.workArea = os.path.join(self.sharedArea, self.workArea)

        # Prepare all the hosts
        for hPar in self.ceParameters["SSHHost"].strip().split(","):
            host = hPar.strip().split("/")[0]
            result = self._prepareRemoteHost(host=host)
            if result["OK"]:
                self.log.info("Host %s registered for usage" % host)
                self.sshHost.append(hPar.strip())
            else:
                self.log.error("Failed to initialize host", host)
                return result

        self.submitOptions = self.ceParameters.get("SubmitOptions", "")
        self.removeOutput = True
        if "RemoveOutput" in self.ceParameters:
            if self.ceParameters["RemoveOutput"].lower() in ["no", "false", "0"]:
                self.removeOutput = False
        self.preamble = self.ceParameters.get("Preamble", "")

        return S_OK()

    #############################################################################
    def submitJob(self, executableFile, proxy, numberOfJobs=1):
        """Method to submit job"""

        # Choose eligible hosts, rank them by the number of available slots
        rankHosts = {}
        maxSlots = 0
        for host in self.sshHost:
            thost = host.split("/")
            hostName = thost[0]
            maxHostJobs = 1
            if len(thost) > 1:
                maxHostJobs = int(thost[1])

            result = self._getHostStatus(hostName)
            if not result["OK"]:
                continue
            slots = maxHostJobs - result["Value"]["Running"]
            if slots > 0:
                rankHosts.setdefault(slots, [])
                rankHosts[slots].append(hostName)
            if slots > maxSlots:
                maxSlots = slots

        if maxSlots == 0:
            return S_ERROR("No online node found on queue")
        # make it executable
        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

        # if no proxy is supplied, the executable can be submitted directly
        # otherwise a wrapper script is needed to get the proxy to the execution node
        # The wrapper script makes debugging more complicated and thus it is
        # recommended to transfer a proxy inside the executable if possible.
        if proxy:
            self.log.verbose("Setting up proxy for payload")
            wrapperContent = bundleProxy(executableFile, proxy)
            name = writeScript(wrapperContent, os.getcwd())
            submitFile = name
        else:  # no proxy
            submitFile = executableFile

        # Submit jobs now
        restJobs = numberOfJobs
        submittedJobs = []
        for slots in range(maxSlots, 0, -1):
            if slots not in rankHosts:
                continue
            for host in rankHosts[slots]:
                result = self._submitJobToHost(submitFile, min(slots, restJobs), host)
                if not result["OK"]:
                    continue
                else:
                    nJobs = len(result["Value"])
                    if nJobs > 0:
                        submittedJobs.extend(result["Value"])
                        restJobs = restJobs - nJobs
                        if restJobs <= 0:
                            break
            if restJobs <= 0:
                break

        if proxy:
            os.remove(submitFile)

        return S_OK(submittedJobs)

    def killJob(self, jobIDs):
        """Kill specified jobs"""
        jobIDList = list(jobIDs)
        if isinstance(jobIDs, six.string_types):
            jobIDList = [jobIDs]

        hostDict = {}
        for job in jobIDList:

            host = os.path.dirname(urlparse(job).path).lstrip("/")
            hostDict.setdefault(host, [])
            hostDict[host].append(job)

        failed = []
        for host, jobIDList in hostDict.items():
            result = self._killJobOnHost(jobIDList, host)
            if not result["OK"]:
                failed.extend(jobIDList)
                message = result["Message"]

        if failed:
            result = S_ERROR(message)
            result["Failed"] = failed
        else:
            result = S_OK()

        return result

    def getCEStatus(self):
        """Method to return information on running and pending jobs."""
        result = S_OK()
        result["SubmittedJobs"] = self.submittedJobs
        result["RunningJobs"] = 0
        result["WaitingJobs"] = 0

        for host in self.sshHost:
            thost = host.split("/")
            resultHost = self._getHostStatus(thost[0])
            if resultHost["OK"]:
                result["RunningJobs"] += resultHost["Value"]["Running"]

        self.log.verbose("Waiting Jobs: ", 0)
        self.log.verbose("Running Jobs: ", result["RunningJobs"])

        return result

    def getJobStatus(self, jobIDList):
        """Get status of the jobs in the given list"""
        hostDict = {}
        for job in jobIDList:
            host = os.path.dirname(urlparse(job).path).lstrip("/")
            hostDict.setdefault(host, [])
            hostDict[host].append(job)

        resultDict = {}
        failed = []
        for host, jobIDList in hostDict.items():
            result = self._getJobStatusOnHost(jobIDList, host)
            if not result["OK"]:
                failed.extend(jobIDList)
                continue
            resultDict.update(result["Value"])

        for job in failed:
            if job not in resultDict:
                resultDict[job] = PilotStatus.UNKNOWN

        return S_OK(resultDict)



"""
:mod: RSSCache

Extension of DictCache to be used within RSS

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import threading
import time

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.DictCache import DictCache


class RSSCache(object):
    """
    Cache with purgeThread integrated
    """

    def __init__(self, lifeTime, updateFunc=None, cacheHistoryLifeTime=None):
        """
        Constructor
        """

        self.__lifeTime = lifeTime
        # lifetime of the history on hours
        self.__cacheHistoryLifeTime = (1 and cacheHistoryLifeTime) or 24
        self.__updateFunc = updateFunc

        # RSSCache
        self.__rssCache = DictCache()
        self.__rssCacheStatus = []  # ( updateTime, message )
        self.__rssCacheLock = threading.Lock()

        # Create purgeThread
        self.__refreshStop = False
        self.__refreshThread = threading.Thread(target=self.__refreshCacheThreadRun)
        self.__refreshThread.setDaemon(True)

    def startRefreshThread(self):
        """
        Run refresh thread.
        """
        self.__refreshThread.start()

    def stopRefreshThread(self):
        """
        Stop refresh thread.
        """
        self.__refreshStop = True

    def isCacheAlive(self):
        """
        Returns status of the cache refreshing thread
        """
        return S_OK(self.__refreshThread.is_alive())

    def setLifeTime(self, lifeTime):
        """
        Set cache life time
        """
        self.__lifeTime = lifeTime

    def setCacheHistoryLifeTime(self, cacheHistoryLifeTime):
        """
        Set cache life time
        """
        self.__cacheHistoryLifeTime = cacheHistoryLifeTime

    def getCacheKeys(self):
        """
        List all the keys stored in the cache.
        """
        self.__rssCacheLock.acquire()
        keys = self.__rssCache.getKeys()
        self.__rssCacheLock.release()

        return S_OK(keys)

    def acquireLock(self):
        """
        Acquires RSSCache lock
        """
        self.__rssCacheLock.acquire()

    def releaseLock(self):
        """
        Releases RSSCache lock
        """
        self.__rssCacheLock.release()

    def getCacheStatus(self):
        """
        Return the latest cache status
        """
        self.__rssCacheLock.acquire()
        if self.__rssCacheStatus:
            res = dict([self.__rssCacheStatus[0]])
        else:
            res = {}
        self.__rssCacheLock.release()
        return S_OK(res)

    def getCacheHistory(self):
        """
        Return the cache updates history
        """
        self.__rssCacheLock.acquire()
        res = dict(self.__rssCacheStatus)
        self.__rssCacheLock.release()
        return S_OK(res)

    def get(self, resourceKey):
        """
        Gets the resource(s) status(es). Every resource can have multiple statuses,
        so in order to speed up things, we store them on the cache as follows::

          { (<resourceName>,<resourceStatusType0>) : whatever0,
            (<resourceName>,<resourceStatusType1>) : whatever1,
          }

        """

        # cacheKey = '%s#%s' % ( resourceName, resourceStatusType )

        self.__rssCacheLock.acquire()
        resourceStatus = self.__rssCache.get(resourceKey)
        self.__rssCacheLock.release()

        if resourceStatus:
            return S_OK({resourceKey: resourceStatus})
        return S_ERROR("Cannot get %s" % resourceKey)

    def getBulk(self, resourceKeys):
        """
        Gets values for resourceKeys in one ATOMIC operation.
        """

        result = {}
        self.__rssCacheLock.acquire()

        for resourceKey in resourceKeys:

            resourceRow = self.__rssCache.get(resourceKey)
            if not resourceRow:
                return S_ERROR("Cannot get %s" % resourceKey)
            result.update({resourceKey: resourceRow})

        self.__rssCacheLock.release()
        return S_OK(result)

    def resetCache(self):
        """
        Reset cache.
        """
        self.__rssCacheLock.acquire()
        self.__rssCache.purgeAll()
        self.__rssCacheLock.release()

        return S_OK()

    def refreshCache(self):
        """
        Clears the cache and gets its latest version, not Thread safe !
        Acquire a lock before using it ! ( and release it afterwards ! )
        """

        self.__rssCache.purgeAll()

        if self.__updateFunc is None:
            return S_ERROR("RSSCache has no updateFunction")
        newCache = self.__updateFunc()
        if not newCache["OK"]:
            return newCache

        itemsAdded = self.__updateCache(newCache["Value"])

        return itemsAdded

    def refreshCacheAndHistory(self):
        """
        Method that refreshes the cache and updates the history. Not thread safe,
        you must acquire a lock before using it, and release it right after !
        """

        refreshResult = self.refreshCache()

        now = datetime.datetime.utcnow()

        if self.__rssCacheStatus:
            # Check oldest record
            dateInserted, _message = self.__rssCacheStatus[-1]
            if dateInserted < now - datetime.timedelta(hours=self.__cacheHistoryLifeTime):
                self.__rssCacheStatus.pop()

        self.__rssCacheStatus.insert(0, (now, refreshResult))

    ################################################################################
    # Private methods

    def __updateCache(self, newCache):
        """
        The new cache must be a dictionary, which should look like::

          { ( <resourceName>,<resourceStatusType0>) : whatever0,
            ( <resourceName>,<resourceStatusType1>) : whatever1,
          }

        """

        itemsCounter = 0

        for cacheKey, cacheValue in newCache.items():
            self.__rssCache.add(cacheKey, self.__lifeTime, value=cacheValue)
            itemsCounter += 1

        return S_OK(itemsCounter)

    def __refreshCacheThreadRun(self):
        """
        Method that refreshes periodically the cache.
        """

        while not self.__refreshStop:

            self.__rssCacheLock.acquire()
            self.refreshCacheAndHistory()
            self.__rssCacheLock.release()

            time.sleep(self.__lifeTime)

        self.__refreshStop = False


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"


class Synchronizer(object):
    """Class encapsulating a lock
    allowing it to be used as a synchronizing
    decorator making the call thread-safe"""

    def __init__(self, lockName="", recursive=False):
        from DIRAC.Core.Utilities.LockRing import LockRing

        self.__lockName = lockName
        self.__lr = LockRing()
        self.__lock = self.__lr.getLock(lockName, recursive=recursive)

    def __call__(self, funcToCall):
        def lockedFunc(*args, **kwargs):
            try:
                if self.__lockName:
                    print("LOCKING", self.__lockName)
                self.__lock.acquire()
                return funcToCall(*args, **kwargs)
            finally:
                if self.__lockName:
                    print("UNLOCKING", self.__lockName)
                self.__lock.release()

        return lockedFunc

    def lock(self):
        return self.__lock.acquire()

    def unlock(self):
        return self.__lock.release()



""" ResourceManagementClient

  Client to interact with the ResourceManagement service and from it with the DB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient


def prepareDict(columnNames, columnValues):
    """
    Convert 2 same size lists into a key->value dict. All Nonetype values are removed.

    :param list columnNames: list containing column names, which are the keys in the returned dict
    :param list columnValues: list of the corresponding values

    :return: dict
    """

    paramsDict = {}

    # make each key name uppercase to match database column names (case sensitive)
    for key, value in zip(columnNames, columnValues):
        if value is not None:
            paramsDict[key] = value

    return paramsDict


@createClient("ResourceStatus/ResourceManagement")
class ResourceManagementClient(Client):
    """
    The :class:`ResourceManagementClient` class exposes the :mod:`DIRAC.ResourceManagement`
    API. All functions you need are on this client.

    You can use this client on this way

     >>> from DIRAC.ResourceManagementSystem.Client.ResourceManagementClient import ResourceManagementClient
     >>> rsClient = ResourceManagementClient()
    """

    def __init__(self, **kwargs):

        super(ResourceManagementClient, self).__init__(**kwargs)
        self.setServer("ResourceStatus/ResourceManagement")

    # AccountingCache Methods ....................................................

    def selectAccountingCache(
        self, name=None, plotType=None, plotName=None, result=None, dateEffective=None, lastCheckTime=None, meta=None
    ):
        """
        Gets from PolicyResult all rows that match the parameters given.

        :param name: name of an individual of the grid topology
        :type name: string, list
        :param plotType: the plotType name (e.g. 'Pilot')
        :type plotType: string, list
        :param plotName: the plot name
        :type plotName: string, list
        :param result: command result
        :type result: string, list
        :param dateEffective: time-stamp from which the result is effective
        :type dateEffective:  datetime, list
        :param lastCheckTime: time-stamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
             For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Name", "PlotType", "PlotName", "Result", "DateEffective", "LastCheckTime", "Meta"]
        columnValues = [name, plotType, plotName, result, dateEffective, lastCheckTime, meta]

        return self._getRPC().select("AccountingCache", prepareDict(columnNames, columnValues))

    def addOrModifyAccountingCache(
        self, name=None, plotType=None, plotName=None, result=None, dateEffective=None, lastCheckTime=None
    ):
        """
        Adds or updates-if-duplicated to AccountingCache. Using `name`, `plotType`
        and `plotName` to query the database, decides whether to insert or update the
        table.

        :param str name: name of an individual of the grid topology
        :param str plotType: name (e.g. 'Pilot')
        :param str plotName: the plot name
        :param str result: command result
        :param datetime dateEffective: timestamp from which the result is effective
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Name", "PlotType", "PlotName", "Result", "DateEffective", "LastCheckTime"]
        columnValues = [name, plotType, plotName, result, dateEffective, lastCheckTime]

        return self._getRPC().addOrModify("AccountingCache", prepareDict(columnNames, columnValues))

    def deleteAccountingCache(
        self, name=None, plotType=None, plotName=None, result=None, dateEffective=None, lastCheckTime=None
    ):
        """
        Deletes from AccountingCache all rows that match the parameters given.

        :param str name: name of an individual of the grid topology
        :param str plotType: the plotType name (e.g. 'Pilot')
        :param str plotName: the plot name
        :param str result: command result
        :param datetime dateEffective: timestamp from which the result is effective
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Name", "PlotType", "PlotName", "Result", "DateEffective", "LastCheckTime"]
        columnValues = [name, plotType, plotName, result, dateEffective, lastCheckTime]

        return self._getRPC().delete("AccountingCache", prepareDict(columnNames, columnValues))

    # GGUSTicketsCache Methods ...................................................

    def selectGGUSTicketsCache(
        self, gocSite=None, link=None, openTickets=None, tickets=None, lastCheckTime=None, meta=None
    ):
        """
        Gets from GGUSTicketsCache all rows that match the parameters given.

        :param str gocSite:
        :param str link: url to the details
        :param int openTickets:
        :param str tickets:
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
           For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """

        columnNames = ["GocSite", "Link", "OpenTickets", "Tickets", "LastCheckTime", "Meta"]
        columnValues = [gocSite, link, openTickets, tickets, lastCheckTime, meta]

        return self._getRPC().select("GGUSTicketsCache", prepareDict(columnNames, columnValues))

    def deleteGGUSTicketsCache(self, gocSite=None, link=None, openTickets=None, tickets=None, lastCheckTime=None):
        """
        Deletes from GGUSTicketsCache all rows that match the parameters given.

        :param str gocSite:
        :param str link: url to the details
        :param int openTickets:
        :param str tickets:
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["GocSite", "Link", "OpenTickets", "Tickets", "LastCheckTime"]
        columnValues = [gocSite, link, openTickets, tickets, lastCheckTime]

        return self._getRPC().delete("GGUSTicketsCache", prepareDict(columnNames, columnValues))

    def addOrModifyGGUSTicketsCache(self, gocSite=None, link=None, openTickets=None, tickets=None, lastCheckTime=None):
        """
        Adds or updates-if-duplicated to GGUSTicketsCache all rows that match the parameters given.

        :param str gocSite:
        :param str link: url to the details
        :param int openTickets:
        :param str tickets:
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["GocSite", "Link", "OpenTickets", "Tickets", "LastCheckTime"]
        columnValues = [gocSite, link, openTickets, tickets, lastCheckTime]

        return self._getRPC().addOrModify("GGUSTicketsCache", prepareDict(columnNames, columnValues))

    # DowntimeCache Methods ......................................................

    def selectDowntimeCache(
        self,
        downtimeID=None,
        element=None,
        name=None,
        startDate=None,
        endDate=None,
        severity=None,
        description=None,
        link=None,
        dateEffective=None,
        lastCheckTime=None,
        gOCDBServiceType=None,
        meta=None,
    ):
        """
        Gets from DowntimeCache all rows that match the parameters given.

        :param downtimeID: unique id for the downtime
        :type downtimeID: string, list
        :param element: valid element in the topology (Site, Resource, Node)
        :type element: string, list
        :param name: name of the element(s) where the downtime applies
        :type name: string, list
        :param startDate: starting time for the downtime
        :type startDate: datetime, list
        :param endDate: ending time for the downtime
        :type endDate: datetime, list
        :param severity: severity assigned by the gocdb
        :type severity: string, list
        :param description: brief description of the downtime
        :type description: string, list
        :param link: url to the details
        :type link: string, list
        :param dateEffective: time when the entry was created in this database
        :type dateEffective: datetime, list
        :param lastCheckTime: timestamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :param str gOCDBServiceType: service type assigned by gocdb
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
            For example: meta={'columns': ['Name']} will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """
        columnNames = [
            "DowntimeID",
            "Element",
            "Name",
            "StartDate",
            "EndDate",
            "Severity",
            "Description",
            "Link",
            "DateEffective",
            "LastCheckTime",
            "GOCDBServiceType",
            "Meta",
        ]
        columnValues = [
            downtimeID,
            element,
            name,
            startDate,
            endDate,
            severity,
            description,
            link,
            dateEffective,
            lastCheckTime,
            gOCDBServiceType,
            meta,
        ]

        return self._getRPC().select("DowntimeCache", prepareDict(columnNames, columnValues))

    def deleteDowntimeCache(
        self,
        downtimeID=None,
        element=None,
        name=None,
        startDate=None,
        endDate=None,
        severity=None,
        description=None,
        link=None,
        dateEffective=None,
        lastCheckTime=None,
        gOCDBServiceType=None,
    ):
        """
        Deletes from DowntimeCache all rows that match the parameters given.

        :param downtimeID: unique id for the downtime
        :type downtimeID: string, list
        :param element: valid element in the topology ( Site, Resource, Node )
        :type element: string, list
        :param name: name of the element where the downtime applies
        :type name: string, list
        :param startDate: starting time for the downtime
        :type startDate: datetime, list
        :param endDate: ending time for the downtime
        :type endDate: datetime, list
        :param severity: severity assigned by the gocdb
        :type severity: string, list
        :param description: brief description of the downtime
        :type description: string, list
        :param link: url to the details
        :type link: string, list
        :param dateEffective: time when the entry was created in this database
        :type dateEffective: datetime, list
        :param lastCheckTime: time-stamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :param str gOCDBServiceType: service type assigned by gocdb
        :return: S_OK() || S_ERROR()
        """
        columnNames = [
            "DowntimeID",
            "Element",
            "Name",
            "StartDate",
            "EndDate",
            "Severity",
            "Description",
            "Link",
            "DateEffective",
            "LastCheckTime",
            "GOCDBServiceType",
        ]
        columnValues = [
            downtimeID,
            element,
            name,
            startDate,
            endDate,
            severity,
            description,
            link,
            dateEffective,
            lastCheckTime,
            gOCDBServiceType,
        ]

        return self._getRPC().delete("DowntimeCache", prepareDict(columnNames, columnValues))

    def addOrModifyDowntimeCache(
        self,
        downtimeID=None,
        element=None,
        name=None,
        startDate=None,
        endDate=None,
        severity=None,
        description=None,
        link=None,
        dateEffective=None,
        lastCheckTime=None,
        gOCDBServiceType=None,
    ):
        """
        Adds or updates-if-duplicated to DowntimeCache. Using `downtimeID` to query
        the database, decides whether to insert or update the table.

        :param str downtimeID: unique id for the downtime
        :param str element: valid element in the topology ( Site, Resource, Node )
        :param str name: name of the element where the downtime applies
        :param datetime startDate: starting time for the downtime
        :param datetime endDate: ending time for the downtime
        :param str severity: severity assigned by the gocdb
        :param str description: brief description of the downtime
        :param str link: url to the details
        :param datetime dateEffective: time when the entry was created in this database
        :param datetime lastCheckTime: timestamp setting last time the result was checked
        :param str gOCDBServiceType: service type assigned by gocdb
        :return: S_OK() || S_ERROR()
        """
        columnNames = [
            "DowntimeID",
            "Element",
            "Name",
            "StartDate",
            "EndDate",
            "Severity",
            "Description",
            "Link",
            "DateEffective",
            "LastCheckTime",
            "GOCDBServiceType",
        ]
        columnValues = [
            downtimeID,
            element,
            name,
            startDate,
            endDate,
            severity,
            description,
            link,
            dateEffective,
            lastCheckTime,
            gOCDBServiceType,
        ]

        return self._getRPC().addOrModify("DowntimeCache", prepareDict(columnNames, columnValues))

    # JobCache Methods ...........................................................

    def selectJobCache(self, site=None, maskStatus=None, efficiency=None, status=None, lastCheckTime=None, meta=None):
        """
        Gets from JobCache all rows that match the parameters given.

        :param site: name of the site element
        :type site: string, list
        :param maskStatus: maskStatus for the site
        :type maskStatus: string, list
        :param efficiency: job efficiency ( successful / total )
        :type efficiency: float, list
        :param status: status for the site computed
        :type status: string, list
        :param lastCheckTime: timestamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
           For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "MaskStatus", "Efficiency", "Status", "LastCheckTime", "Meta"]
        columnValues = [site, maskStatus, efficiency, status, lastCheckTime, meta]

        return self._getRPC().select("JobCache", prepareDict(columnNames, columnValues))

    def deleteJobCache(self, site=None, maskStatus=None, efficiency=None, status=None, lastCheckTime=None):
        """
        Deletes from JobCache all rows that match the parameters given.

        :param site: name of the site element
        :type site: string, list
        :param maskStatus: maskStatus for the site
        :type maskStatus: string, list
        :param efficiency: job efficiency ( successful / total )
        :type efficiency: float, list
        :param status: status for the site computed
        :type status: string, list
        :param lastCheckTime: timestamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "MaskStatus", "Efficiency", "Status", "LastCheckTime"]
        columnValues = [site, maskStatus, efficiency, status, lastCheckTime]

        return self._getRPC().delete("JobCache", prepareDict(columnNames, columnValues))

    def addOrModifyJobCache(self, site=None, maskStatus=None, efficiency=None, status=None, lastCheckTime=None):
        """
        Adds or updates-if-duplicated to JobCache. Using `site` to query
        the database, decides whether to insert or update the table.

        :param site: name of the site element
        :type site: string, list
        :param maskStatus: maskStatus for the site
        :type maskStatus: string, list
        :param efficiency: job efficiency ( successful / total )
        :type efficiency: float, list
        :param status: status for the site computed
        :type status: string, list
        :param lastCheckTime: time-stamp setting last time the result was checked
        :type lastCheckTime: datetime, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "MaskStatus", "Efficiency", "Status", "LastCheckTime"]
        columnValues = [site, maskStatus, efficiency, status, lastCheckTime]

        return self._getRPC().addOrModify("JobCache", prepareDict(columnNames, columnValues))

    # TransferCache Methods ......................................................

    def selectTransferCache(
        self, sourceName=None, destinationName=None, metric=None, value=None, lastCheckTime=None, meta=None
    ):
        """
        Gets from TransferCache all rows that match the parameters given.

        :param elementName: name of the element
        :type elementName: string, list
        :param direction: the element taken as Source or Destination of the transfer
        :type direction: string, list
        :param metric: measured quality of failed transfers
        :type metric: string, list
        :param value: percentage
        :type value: float, list
        :param lastCheckTime: time-stamp setting last time the result was checked
        :type lastCheckTime: float, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
          For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["SourceName", "DestinationName", "Metric", "Value", "LastCheckTime", "Meta"]
        columnValues = [sourceName, destinationName, metric, value, lastCheckTime, meta]

        return self._getRPC().select("TransferCache", prepareDict(columnNames, columnValues))

    def deleteTransferCache(self, sourceName=None, destinationName=None, metric=None, value=None, lastCheckTime=None):
        """
         Deletes from TransferCache all rows that match the parameters given.

        :param elementName: name of the element
        :type elementName: string, list
        :param direction: the element taken as Source or Destination of the transfer
        :type direction: string, list
        :param metric: measured quality of failed transfers
        :type metric: string, list
        :param value: percentage
        :type value: float, list
        :param lastCheckTime: time-stamp setting last time the result was checked
        :type lastCheckTime: float, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["SourceName", "DestinationName", "Metric", "Value", "LastCheckTime"]
        columnValues = [sourceName, destinationName, metric, value, lastCheckTime]

        return self._getRPC().delete("TransferCache", prepareDict(columnNames, columnValues))

    def addOrModifyTransferCache(
        self, sourceName=None, destinationName=None, metric=None, value=None, lastCheckTime=None
    ):
        """
         Adds or updates-if-duplicated to TransferCache. Using `elementName`, `direction`
         and `metric` to query the database, decides whether to insert or update the table.

        :param str elementName: name of the element
        :param str direction: the element taken as Source or Destination of the transfer
        :param str metric: measured quality of failed transfers
        :param float value: percentage
        :param datetime lastCheckTime: time-stamp setting last time the result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["SourceName", "DestinationName", "Metric", "Value", "LastCheckTime"]
        columnValues = [sourceName, destinationName, metric, value, lastCheckTime]

        return self._getRPC().addOrModify("TransferCache", prepareDict(columnNames, columnValues))

    # PilotCache Methods .........................................................

    def selectPilotCache(
        self,
        site=None,
        cE=None,
        pilotsPerJob=None,
        pilotJobEff=None,
        status=None,
        lastCheckTime=None,
        meta=None,
        vO=None,
    ):
        """
        Gets from TransferCache all rows that match the parameters given.

        :param site: name of the site
        :type site: string, list
        :param cE: name of the CE of 'Multiple' if all site CEs are considered
        :type cE: string, list
        :param pilotsPerJob: measure calculated
        :type pilotsPerJob: float, list
        :param pilotJobEff: percentage
        :type pilotJobEff: float, list
        :param status: status of the CE / Site
        :type status: float, list
        :param lastCheckTime: measure calculated
        :type lastCheckTime: datetime, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
           For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "CE", "PilotsPerJob", "PilotJobEff", "Status", "LastCheckTime", "Meta", "VO"]
        columnValues = [site, cE, pilotsPerJob, pilotJobEff, status, lastCheckTime, meta, vO]

        return self._getRPC().select("PilotCache", prepareDict(columnNames, columnValues))

    def deletePilotCache(
        self, site=None, cE=None, pilotsPerJob=None, pilotJobEff=None, status=None, lastCheckTime=None, vO=None
    ):
        """
        Deletes from TransferCache all rows that match the parameters given.

        :param site: name of the site
        :type site: string, list
        :param cE: name of the CE of 'Multiple' if all site CEs are considered
        :type cE: string, list
        :param pilotsPerJob: measure calculated
        :type pilotsPerJob: float, list
        :param pilotJobEff: percentage
        :type pilotJobEff: float, list
        :param status: status of the CE / Site
        :type status: float, list
        :param lastCheckTime: measure calculated
        :type lastCheckTime: datetime, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "CE", "PilotsPerJob", "PilotJobEff", "Status", "LastCheckTime", "VO"]
        columnValues = [site, cE, pilotsPerJob, pilotJobEff, status, lastCheckTime, vO]

        return self._getRPC().delete("PilotCache", prepareDict(columnNames, columnValues))

    def addOrModifyPilotCache(
        self, site=None, cE=None, pilotsPerJob=None, pilotJobEff=None, status=None, lastCheckTime=None, vO=None
    ):
        """
        Adds or updates-if-duplicated to PilotCache. Using `site` and `cE`
        to query the database, decides whether to insert or update the table.

        :param str site: name of the site
        :param str cE: name of the CE of 'Multiple' if all site CEs are considered
        :param float pilotsPerJob: measure calculated
        :param flaot pilotJobEff: percentage
        :param str status: status of the CE / Site
        :param datetime lastCheckTime: measure calculated
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Site", "CE", "PilotsPerJob", "PilotJobEff", "Status", "LastCheckTime", "VO"]
        columnValues = [site, cE, pilotsPerJob, pilotJobEff, status, lastCheckTime, vO]

        return self._getRPC().addOrModify("PilotCache", prepareDict(columnNames, columnValues))

    # PolicyResult Methods .......................................................

    def selectPolicyResult(
        self,
        element=None,
        name=None,
        policyName=None,
        statusType=None,
        status=None,
        reason=None,
        lastCheckTime=None,
        meta=None,
        vO=None,
    ):
        """
        Gets from PolicyResult all rows that match the parameters given.

        :param granularity: it has to be a valid element ( ValidElement ), any of the defaults:
           'Site' | 'Service' | 'Resource' | 'StorageElement'
        :type granularity: string, list
        :param name: name of the element
        :type name: string, list
        :param policyName: name of the policy
        :type policyName: string, list
        :param statusType: it has to be a valid status type for the given granularity
        :type statusType: string, list
        :param status: it has to be a valid status, any of the defaults:
            'Active' | 'Degraded' | 'Probing' | 'Banned'
        :type status: string, list
        :param reason: decision that triggered the assigned status
        :type reason: string, list
        :param lastCheckTime: time-stamp setting last time the policy result was checked
        :type lastCheckTime: datetime, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
            For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Element", "Name", "PolicyName", "StatusType", "Status", "Reason", "LastCheckTime", "Meta", "VO"]
        columnValues = [element, name, policyName, statusType, status, reason, lastCheckTime, meta, vO]

        return self._getRPC().select("PolicyResult", prepareDict(columnNames, columnValues))

    def deletePolicyResult(
        self,
        element=None,
        name=None,
        policyName=None,
        statusType=None,
        status=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        vO=None,
    ):
        """
        Deletes from PolicyResult all rows that match the parameters given.

        :param granularity: it has to be a valid element ( ValidElement ), any of the defaults:
           'Site' | 'Service' | 'Resource' | 'StorageElement'
        :type granularity: string, list
        :param name: name of the element
        :type name: string, list
        :param policyName: name of the policy
        :type policyName: string, list
        :param statusType: it has to be a valid status type for the given granularity
        :type statusType: string, list
        :param status: it has to be a valid status, any of the defaults: 'Active' | 'Degraded' | 'Probing' | 'Banned'
        :type status: string, list
        :param reason: decision that triggered the assigned status
        :type reason: string, list
        :param datetime dateEffective: time-stamp from which the policy result is effective
        :param lastCheckTime: time-stamp setting last time the policy result was checked
        :type lastCheckTime: datetime, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = [
            "Element",
            "Name",
            "PolicyName",
            "StatusType",
            "Status",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "VO",
        ]
        columnValues = [element, name, policyName, statusType, status, reason, dateEffective, lastCheckTime, vO]

        return self._getRPC().delete("PolicyResult", prepareDict(columnNames, columnValues))

    def addOrModifyPolicyResult(
        self,
        element=None,
        name=None,
        policyName=None,
        statusType=None,
        status=None,
        reason=None,
        dateEffective=None,
        lastCheckTime=None,
        vO=None,
    ):
        """
        Adds or updates-if-duplicated to PolicyResult. Using `name`, `policyName` and
        `statusType` to query the database, decides whether to insert or update the table.

        :param str element: it has to be a valid element ( ValidElement ), any of the defaults:
           'Site' | 'Service' | 'Resource' | 'StorageElement'
        :param str name: name of the element
        :param str policyName: name of the policy
        :param str statusType: it has to be a valid status type for the given element
        :param str status: it has to be a valid status, any of the defaults:
          'Active' | 'Degraded' | 'Probing' | 'Banned'
        :param str reason: decision that triggered the assigned status
        :param datetime dateEffective: time-stamp from which the policy result is effective
        :param datetime lastCheckTime: time-stamp setting last time the policy result was checked
        :return: S_OK() || S_ERROR()
        """
        columnNames = [
            "Element",
            "Name",
            "PolicyName",
            "StatusType",
            "Status",
            "Reason",
            "DateEffective",
            "LastCheckTime",
            "VO",
        ]
        columnValues = [element, name, policyName, statusType, status, reason, dateEffective, lastCheckTime, vO]

        return self._getRPC().addOrModify("PolicyResult", prepareDict(columnNames, columnValues))

    # SpaceTokenOccupancyCache Methods ...........................................

    def selectSpaceTokenOccupancyCache(
        self, endpoint=None, token=None, total=None, guaranteed=None, free=None, lastCheckTime=None, meta=None
    ):
        """
        Gets from SpaceTokenOccupancyCache all rows that match the parameters given.

        :param endpoint: endpoint
        :type endpoint: string, list
        :param token: name of the token
        :type token: string, list
        :param total: total terabytes
        :type total: integer, list
        :param guaranteed: guaranteed terabytes
        :type guaranteed: integer, list
        :param free: free terabytes
        :type free: integer, list
        :param lastCheckTime: time-stamp from which the result is effective
        :type lastCheckTime: datetime, list
        :param dict meta: metadata for the mysql query. Currently it is being used only for column selection.
            For example: meta={'columns': ['Name']} will return only the 'Name' column.
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Endpoint", "Token", "Total", "Guaranteed", "Free", "LastCheckTime", "Meta"]
        columnValues = [endpoint, token, total, guaranteed, free, lastCheckTime, meta]

        return self._getRPC().select("SpaceTokenOccupancyCache", prepareDict(columnNames, columnValues))

    def deleteSpaceTokenOccupancyCache(
        self, endpoint=None, token=None, total=None, guaranteed=None, free=None, lastCheckTime=None
    ):
        """
        Deletes from SpaceTokenOccupancyCache all rows that match the parameters given.

        :param endpoint: endpoint
        :type endpoint: string, list
        :param token: name of the token
        :type token: string, list
        :param total: total terabytes
        :type total: integer, list
        :param guaranteed: guaranteed terabytes
        :type guaranteed: integer, list
        :param free: free terabytes
        :type free: integer, list
        :param lastCheckTime: time-stamp from which the result is effective
        :type lastCheckTime: datetime, list
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Endpoint", "Token", "Total", "Guaranteed", "Free", "LastCheckTime"]
        columnValues = [endpoint, token, total, guaranteed, free, lastCheckTime]

        return self._getRPC().delete("SpaceTokenOccupancyCache", prepareDict(columnNames, columnValues))

    def addOrModifySpaceTokenOccupancyCache(
        self, endpoint=None, token=None, total=None, guaranteed=None, free=None, lastCheckTime=None
    ):
        """
        Adds or updates-if-duplicated to SpaceTokenOccupancyCache. Using `site` and `token`
        to query the database, decides whether to insert or update the table.

        :param endpoint: endpoint
        :type endpoint: string, list
        :param str token: name of the token
        :param int total: total terabytes
        :param int guaranteed: guaranteed terabytes
        :param int free: free terabytes
        :param datetime lastCheckTime: time-stamp from which the result is effective
        :return: S_OK() || S_ERROR()
        """
        columnNames = ["Endpoint", "Token", "Total", "Guaranteed", "Free", "LastCheckTime"]
        columnValues = [endpoint, token, total, guaranteed, free, lastCheckTime]

        return self._getRPC().addOrModify("SpaceTokenOccupancyCache", prepareDict(columnNames, columnValues))


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



# pylint: disable=protected-access
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import tempfile

from DIRAC.Interfaces.API.Job import Job
from DIRAC.tests.Utilities.utils import find_all


def helloWorldJob():
    job = Job()
    job.setName("helloWorld")
    exeScriptLocation = find_all("exe-script.py", "..", "/DIRAC/tests/Integration")[0]
    job.setInputSandbox(exeScriptLocation)
    job.setExecutable(exeScriptLocation, "", "helloWorld.log")
    return job


def parametricJob():
    job = Job()
    job.setName("parametric_helloWorld_%n")
    exeScriptLocation = find_all("exe-script.py", "..", "/DIRAC/tests/Integration")[0]
    job.setInputSandbox(exeScriptLocation)
    job.setParameterSequence("args", ["one", "two", "three"])
    job.setParameterSequence("iargs", [1, 2, 3])
    job.setExecutable(exeScriptLocation, arguments=": testing %(args)s %(iargs)s", logFile="helloWorld_%n.log")
    return job


def createFile(job):
    tmpdir = tempfile.mkdtemp()
    jobDescription = tmpdir + "/jobDescription.xml"
    with open(jobDescription, "w") as fd:
        fd.write(job._toXML())
    return jobDescription



#!/usr/bin/env python
########################################################################
# File :    dirac-utils-file-md5
# Author :
########################################################################
"""
Calculate md5 of the supplied file

Example:
  $ dirac-utils-file-md5 Example.tgz
  Example.tgz 5C1A1102-EAFD-2CBA-25BD-0EFCCFC3623E
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["File:     File Name"])
    _, files = Script.parseCommandLine(ignoreErrors=False)

    exitCode = 0

    import DIRAC
    from DIRAC.Core.Utilities.File import makeGuid

    for file in files:
        try:
            md5 = makeGuid(file)
            if md5:
                print(file.rjust(100), md5.ljust(10))
            else:
                print("ERROR %s: Failed to get md5" % file)
                exitCode = 2
        except Exception as x:
            print("ERROR %s: Failed to get md5" % file, str(x))
            exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Helper for /Registry section
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import errno

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import getVO

__RCSID__ = "$Id$"

# pylint: disable=missing-docstring

gBaseRegistrySection = "/Registry"


def getUsernameForDN(dn, usersList=None):
    """Find DIRAC user for DN

    :param str dn: user DN
    :param list usersList: list of possible users

    :return: S_OK(str)/S_ERROR()
    """
    dn = dn.strip()
    if not usersList:
        result = gConfig.getSections("%s/Users" % gBaseRegistrySection)
        if not result["OK"]:
            return result
        usersList = result["Value"]
    for username in usersList:
        if dn in gConfig.getValue("%s/Users/%s/DN" % (gBaseRegistrySection, username), []):
            return S_OK(username)
    return S_ERROR("No username found for dn %s" % dn)


def getDNForUsername(username):
    """Get user DN for user

    :param str username: user name

    :return: S_OK(str)/S_ERROR()
    """
    dnList = gConfig.getValue("%s/Users/%s/DN" % (gBaseRegistrySection, username), [])
    return S_OK(dnList) if dnList else S_ERROR("No DN found for user %s" % username)


def getDNForHost(host):
    """Get host DN

    :param str host: host domain

    :return: S_OK(list)/S_ERROR() -- list of DNs
    """
    dnList = gConfig.getValue("%s/Hosts/%s/DN" % (gBaseRegistrySection, host), [])
    return S_OK(dnList) if dnList else S_ERROR("No DN found for host %s" % host)


def getGroupsForDN(dn):
    """Get all possible groups for DN

    :param str dn: user DN

    :return: S_OK(list)/S_ERROR() -- contain list of groups
    """
    dn = dn.strip()
    result = getUsernameForDN(dn)
    if not result["OK"]:
        return result
    return getGroupsForUser(result["Value"])


def __getGroupsWithAttr(attrName, value):
    """Get all possible groups with some attribute

    :param str attrName: attribute name
    :param str value: attribute value

    :return: S_OK(list)/S_ERROR() -- contain list of groups
    """
    result = gConfig.getSections("%s/Groups" % gBaseRegistrySection)
    if not result["OK"]:
        return result
    groupsList = result["Value"]
    groups = []
    for group in groupsList:
        if value in gConfig.getValue("%s/Groups/%s/%s" % (gBaseRegistrySection, group, attrName), []):
            groups.append(group)
    groups.sort()
    return S_OK(groups) if groups else S_ERROR("No groups found for %s=%s" % (attrName, value))


def getGroupsForUser(username):
    """Find groups for user

    :param str username: user name

    :return: S_OK(list)/S_ERROR() -- contain list of groups
    """
    return __getGroupsWithAttr("Users", username)


def getGroupsForVO(vo):
    """Get groups for VO

    :param str vo: DIRAC VO name

    :return: S_OK(list)/S_ERROR()
    """
    if getVO():  # tries to get default VO in /DIRAC/VirtualOrganization
        return gConfig.getSections("%s/Groups" % gBaseRegistrySection)
    if not vo:
        return S_ERROR("No VO requested")
    return __getGroupsWithAttr("VO", vo)


def getGroupsWithProperty(propName):
    """Search groups by property

    :param str propName: property name

    :return: S_OK(list)/S_ERROR()
    """
    return __getGroupsWithAttr("Properties", propName)


def getHostnameForDN(dn):
    """Search host name for host DN

    :param str dn: host DN

    :return: S_OK()/S_ERROR()
    """
    dn = dn.strip()
    result = gConfig.getSections("%s/Hosts" % gBaseRegistrySection)
    if not result["OK"]:
        return result
    hostList = result["Value"]
    for hostname in hostList:
        if dn in gConfig.getValue("%s/Hosts/%s/DN" % (gBaseRegistrySection, hostname), []):
            return S_OK(hostname)
    return S_ERROR("No hostname found for dn %s" % dn)


def getDefaultUserGroup():
    """Search general default group

    :return: str
    """
    return gConfig.getValue("/%s/DefaultGroup" % gBaseRegistrySection, "user")


def findDefaultGroupForDN(dn):
    """Search defaut group for DN

    :param str dn: DN

    :return: S_OK()/S_ERROR()
    """
    dn = dn.strip()
    result = getUsernameForDN(dn)
    if not result["OK"]:
        return result
    return findDefaultGroupForUser(result["Value"])


def findDefaultGroupForUser(userName):
    """Get default group for user

    :param str userName: user name

    :return: S_OK(str)/S_ERROR()
    """
    defGroups = getUserOption(userName, "DefaultGroup", [])
    defGroups += gConfig.getValue("%s/DefaultGroup" % gBaseRegistrySection, ["user"])
    result = getGroupsForUser(userName)
    if not result["OK"]:
        return result
    userGroups = result["Value"]
    for group in defGroups:
        if group in userGroups:
            return S_OK(group)
    return S_OK(userGroups[0]) if userGroups else S_ERROR("User %s has no groups" % userName)


def getAllUsers():
    """Get all users

    :return: list
    """
    result = gConfig.getSections("%s/Users" % gBaseRegistrySection)
    return result["Value"] if result["OK"] else []


def getAllGroups():
    """Get all groups

    :return: list
    """
    result = gConfig.getSections("%s/Groups" % gBaseRegistrySection)
    return result["Value"] if result["OK"] else []


def getUsersInGroup(groupName, defaultValue=None):
    """Find all users for group

    :param str group: group name
    :param defaultValue: default value

    :return: list
    """
    option = "%s/Groups/%s/Users" % (gBaseRegistrySection, groupName)
    return gConfig.getValue(option, [] if defaultValue is None else defaultValue)


def getUsersInVO(vo, defaultValue=None):
    """Search users in VO

    :param str vo: DIRAC VO name
    :param defaultValue: default value

    :return: list
    """
    result = getGroupsForVO(vo)
    if not result["OK"] or not result["Value"]:
        return [] if defaultValue is None else defaultValue
    groups = result["Value"]

    userList = []
    for group in groups:
        userList += getUsersInGroup(group)
    return userList


def getDNsInVO(vo):
    """Get all DNs that have a VO users

    :param str vo: VO name

    :return: list
    """
    DNs = []
    for user in getUsersInVO(vo):
        result = getDNForUsername(user)
        if result["OK"]:
            DNs.extend(result["Value"])
    return DNs


def getDNsInGroup(groupName):
    """Find all DNs  for DIRAC group

    :param str groupName: group name

    :return: list
    """
    DNs = []
    for user in getUsersInGroup(groupName):
        result = getDNForUsername(user)
        if result["OK"]:
            DNs.extend(result["Value"])
    return DNs


def getPropertiesForGroup(groupName, defaultValue=None):
    """Return group properties

    :param str groupName: group name
    :param defaultValue: default value

    :return: defaultValue or list
    """
    option = "%s/Groups/%s/Properties" % (gBaseRegistrySection, groupName)
    return gConfig.getValue(option, [] if defaultValue is None else defaultValue)


def getPropertiesForHost(hostName, defaultValue=None):
    """Return host properties

    :param str hostName: host name
    :param defaultValue: default value

    :return: defaultValue or list
    """
    option = "%s/Hosts/%s/Properties" % (gBaseRegistrySection, hostName)
    return gConfig.getValue(option, [] if defaultValue is None else defaultValue)


def getPropertiesForEntity(group, name="", dn="", defaultValue=None):
    """Return some entity properties

    :param str group: group name
    :param str name: entity name
    :param str dn: DN
    :param defaultValue: default value

    :return: defaultValue or list
    """
    if defaultValue is None:
        defaultValue = []
    if group == "hosts":
        if not name:
            result = getHostnameForDN(dn)
            if not result["OK"]:
                return [] if defaultValue is None else defaultValue
            name = result["Value"]
        return getPropertiesForHost(name, defaultValue)
    return getPropertiesForGroup(group, defaultValue)


def __matchProps(sProps, rProps):
    """Match properties

    :param sProps: submitted properties
    :param rProps: required properties

    :return: list -- contain matched properties
    """
    foundProps = []
    for prop in sProps:
        if prop in rProps:
            foundProps.append(prop)
    return foundProps


def groupHasProperties(groupName, propList):
    """Match required properties with group properties

    :param str groupName: group name
    :param list propList: required properties

    :return: list -- contain matched properties
    """

    if isinstance(propList, six.string_types):
        propList = [propList]
    return __matchProps(propList, getPropertiesForGroup(groupName))


def hostHasProperties(hostName, propList):
    """Match required properties with host properties

    :param str hostName: host name
    :param list propList: required properties

    :return: list -- contain matched properties
    """
    if isinstance(propList, six.string_types):
        propList = [propList]
    return __matchProps(propList, getPropertiesForHost(hostName))


def getUserOption(userName, optName, defaultValue=""):
    """Get user option

    :param str userName: user name
    :param str optName: option name
    :param defaultValue: default value

    :return: defaultValue or str
    """
    return gConfig.getValue("%s/Users/%s/%s" % (gBaseRegistrySection, userName, optName), defaultValue)


def getGroupOption(groupName, optName, defaultValue=""):
    """Get group option

    :param str groupName: group name
    :param str optName: option name
    :param defaultValue: default value

    :return: defaultValue or str
    """
    return gConfig.getValue("%s/Groups/%s/%s" % (gBaseRegistrySection, groupName, optName), defaultValue)


def getHostOption(hostName, optName, defaultValue=""):
    """Get host option

    :param str hostName: host name
    :param str optName: option name
    :param defaultValue: default value

    :return: defaultValue or str
    """
    return gConfig.getValue("%s/Hosts/%s/%s" % (gBaseRegistrySection, hostName, optName), defaultValue)


def getHosts():
    """Get all hosts

    :return: S_OK()/S_ERROR()
    """
    return gConfig.getSections("%s/Hosts" % gBaseRegistrySection)


def getVOOption(voName, optName, defaultValue=""):
    """Get VO option

    :param str voName: DIRAC VO name
    :param str optName: option name
    :param defaultValue: default value

    :return: defaultValue or str
    """
    return gConfig.getValue("%s/VO/%s/%s" % (gBaseRegistrySection, voName, optName), defaultValue)


def getBannedIPs():
    """Get banned IPs

    :return: list
    """
    return gConfig.getValue("%s/BannedIPs" % gBaseRegistrySection, [])


def getVOForGroup(group):
    """Search VO name for group

    :param str group: group name

    :return: str
    """
    return getVO() or gConfig.getValue("%s/Groups/%s/VO" % (gBaseRegistrySection, group), "")


def getDefaultVOMSAttribute():
    """Get default VOMS attribute

    :return: str
    """
    return gConfig.getValue("%s/DefaultVOMSAttribute" % gBaseRegistrySection, "")


def getVOMSAttributeForGroup(group):
    """Search VOMS attribute for group

    :param str group: group name

    :return: str
    """
    return gConfig.getValue("%s/Groups/%s/VOMSRole" % (gBaseRegistrySection, group), getDefaultVOMSAttribute())


def getDefaultVOMSVO():
    """Get default VOMS VO

    :return: str
    """
    return gConfig.getValue("%s/DefaultVOMSVO" % gBaseRegistrySection, "") or getVO()


def getVOMSVOForGroup(group):
    """Search VOMS VO for group

    :param str group: group name

    :return: str
    """
    vomsVO = gConfig.getValue("%s/Groups/%s/VOMSVO" % (gBaseRegistrySection, group), getDefaultVOMSVO())
    if not vomsVO:
        vo = getVOForGroup(group)
        vomsVO = getVOOption(vo, "VOMSName", "")
    return vomsVO


def getGroupsWithVOMSAttribute(vomsAttr):
    """Search groups with VOMS attribute

    :param str vomsAttr: VOMS attribute

    :return: list
    """
    groups = []
    for group in gConfig.getSections("%s/Groups" % (gBaseRegistrySection)).get("Value", []):
        if vomsAttr == gConfig.getValue("%s/Groups/%s/VOMSRole" % (gBaseRegistrySection, group), ""):
            groups.append(group)
    return groups


def getVOs():
    """Get all the configured VOs

    :return: S_OK(list)/S_ERROR()
    """
    voName = getVO()
    return S_OK([voName]) if voName else gConfig.getSections("%s/VO" % gBaseRegistrySection)


def getVOMSServerInfo(requestedVO=""):
    """Get information on VOMS servers for the given VO or for all of them

    :param str requestedVO: requested VO

    :return: S_OK()/S_ERROR()
    """
    vomsDict = {}

    result = getVOs()
    if result["OK"]:
        voNames = result["Value"]
        for vo in voNames:
            if requestedVO and vo != requestedVO:
                continue
            vomsName = getVOOption(vo, "VOMSName", "")
            if not vomsName:
                continue
            vomsDict.setdefault(vo, {})
            vomsDict[vo]["VOMSName"] = getVOOption(vo, "VOMSName", "")
            result = gConfig.getSections("%s/VO/%s/VOMSServers" % (gBaseRegistrySection, vo))
            if result["OK"]:
                serverList = result["Value"]
                vomsDict[vo].setdefault("Servers", {})
                for server in serverList:
                    vomsDict[vo]["Servers"].setdefault(server, {})
                    DN = gConfig.getValue("%s/VO/%s/VOMSServers/%s/DN" % (gBaseRegistrySection, vo, server), "")
                    CA = gConfig.getValue("%s/VO/%s/VOMSServers/%s/CA" % (gBaseRegistrySection, vo, server), "")
                    port = gConfig.getValue("%s/VO/%s/VOMSServers/%s/Port" % (gBaseRegistrySection, vo, server), 0)
                    vomsDict[vo]["Servers"][server]["DN"] = DN
                    vomsDict[vo]["Servers"][server]["CA"] = CA
                    vomsDict[vo]["Servers"][server]["Port"] = port

    return S_OK(vomsDict)


def getVOMSRoleGroupMapping(vo=""):
    """Get mapping of the VOMS role to the DIRAC group

    :param str vo: perform the operation for the given VO

    :return: S_OK(dict)/S_ERROR() -- dictionary have standard structure with two mappings:
             VOMS-DIRAC { <VOMS_Role>: [<DIRAC_Group>] },
             DIRAC-VOMS { <DIRAC_Group>: <VOMS_Role> }
             and a list of DIRAC groups without mapping
    """
    result = getGroupsForVO(vo)
    if not result["OK"]:
        return result

    groupList = result["Value"]

    vomsGroupDict = {}
    groupVomsDict = {}
    noVOMSGroupList = []
    noVOMSSyncGroupList = []

    for group in groupList:
        vomsRole = getGroupOption(group, "VOMSRole")
        if vomsRole:
            vomsGroupDict.setdefault(vomsRole, [])
            vomsGroupDict[vomsRole].append(group)
            groupVomsDict[group] = vomsRole
            syncVOMS = getGroupOption(group, "AutoSyncVOMS", True)
            if not syncVOMS:
                noVOMSSyncGroupList.append(group)

    for group in groupList:
        if group not in groupVomsDict:
            noVOMSGroupList.append(group)

    return S_OK(
        {
            "VOMSDIRAC": vomsGroupDict,
            "DIRACVOMS": groupVomsDict,
            "NoVOMS": noVOMSGroupList,
            "NoSyncVOMS": noVOMSSyncGroupList,
        }
    )


def getUsernameForID(ID, usersList=None):
    """Get DIRAC user name by ID

    :param str ID: user ID
    :param list usersList: list of DIRAC user names

    :return: S_OK(str)/S_ERROR()
    """
    if not usersList:
        result = gConfig.getSections("%s/Users" % gBaseRegistrySection)
        if not result["OK"]:
            return result
        usersList = result["Value"]
    for username in usersList:
        if ID in gConfig.getValue("%s/Users/%s/ID" % (gBaseRegistrySection, username), []):
            return S_OK(username)
    return S_ERROR("No username found for ID %s" % ID)


def getCAForUsername(username):
    """Get CA option by user name

    :param str username: user name

    :return: S_OK(str)/S_ERROR()
    """
    dnList = gConfig.getValue("%s/Users/%s/CA" % (gBaseRegistrySection, username), [])
    return S_OK(dnList) if dnList else S_ERROR("No CA found for user %s" % username)


def getDNProperty(userDN, value, defaultValue=None):
    """Get property from DNProperties section by user DN

    :param str userDN: user DN
    :param str value: option that need to get
    :param defaultValue: default value

    :return: S_OK()/S_ERROR() -- str or list that contain option value
    """
    result = getUsernameForDN(userDN)
    if not result["OK"]:
        return result
    pathDNProperties = "%s/Users/%s/DNProperties" % (gBaseRegistrySection, result["Value"])
    result = gConfig.getSections(pathDNProperties)
    if result["OK"]:
        for section in result["Value"]:
            if userDN == gConfig.getValue("%s/%s/DN" % (pathDNProperties, section)):
                return S_OK(gConfig.getValue("%s/%s/%s" % (pathDNProperties, section, value), defaultValue))
    return S_OK(defaultValue)


def getProxyProvidersForDN(userDN):
    """Get proxy providers by user DN

    :param str userDN: user DN

    :return: S_OK(list)/S_ERROR()
    """
    return getDNProperty(userDN, "ProxyProviders", [])


def getDNFromProxyProviderForUserID(proxyProvider, userID):
    """Get groups by user DN in DNProperties

    :param str proxyProvider: proxy provider name
    :param str userID: user identificator

    :return: S_OK(str)/S_ERROR()
    """
    # Get user name
    result = getUsernameForID(userID)
    if not result["OK"]:
        return result
    # Get DNs from user
    result = getDNForUsername(result["Value"])
    if not result["OK"]:
        return result
    for DN in result["Value"]:
        result = getProxyProvidersForDN(DN)
        if not result["OK"]:
            return result
        if proxyProvider in result["Value"]:
            return S_OK(DN)
    return S_ERROR(errno.ENODATA, "No DN found for %s proxy provider for user ID %s" % (proxyProvider, userID))


def isDownloadableGroup(groupName):
    """Get permission to download proxy with group in a argument

    :params str groupName: DIRAC group

    :return: boolean
    """
    if getGroupOption(groupName, "DownloadableProxy") in [False, "False", "false", "no"]:
        return False
    return True


def getUserDict(username):
    """Get full information from user section

    :param str username: DIRAC user name

    :return: S_OK()/S_ERROR()
    """
    resDict = {}
    relPath = "%s/Users/%s/" % (gBaseRegistrySection, username)
    result = gConfig.getConfigurationTree(relPath)
    if not result["OK"]:
        return result
    for key, value in result["Value"].items():
        if value:
            resDict[key.replace(relPath, "")] = value
    return S_OK(resDict)


def getEmailsForGroup(groupName):
    """Get email list of users in group

    :param str groupName: DIRAC group name

    :return: list(list) -- inner list contains emails for a user
    """
    emails = []
    for username in getUsersInGroup(groupName):
        email = getUserOption(username, "Email", [])
        emails.append(email)
    return emails



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import random
import time
import threading
import six
from six.moves import _thread as thread
from hashlib import md5

from DIRAC.Core.Utilities.ReturnValues import S_ERROR, S_OK
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton

__RCSID__ = "$Id$"


@six.add_metaclass(DIRACSingleton)
class LockRing(object):
    def __init__(self):
        self.__locks = {}
        self.__events = {}

    def __genName(self, container):
        # TODO: Shouldn't this be a UUID?
        name = md5(str(time.time() + random.random()).encode()).hexdigest()
        retries = 10
        while name in container and retries:
            name = md5(str(time.time() + random.random()).encode()).hexdigest()
            retries -= 1
        return name

    def getLock(self, lockName="", recursive=False):
        if not lockName:
            lockName = self.__genName(self.__locks)
        try:
            return self.__locks[lockName]
        except KeyError:
            if recursive:
                self.__locks[lockName] = threading.RLock()
            else:
                self.__locks[lockName] = threading.Lock()
        return self.__locks[lockName]

    def getEvent(self, evName=""):
        if not evName:
            evName = self.__genName(self.__events)
        try:
            return self.__events[evName]
        except KeyError:
            self.__events[evName] = threading.Event()
        return self.__events[evName]

    def acquire(self, lockName):
        try:
            self.__locks[lockName].acquire()
        except ValueError:
            return S_ERROR("No lock named %s" % lockName)
        return S_OK()

    def release(self, lockName):
        try:
            self.__locks[lockName].release()
        except ValueError:
            return S_ERROR("No lock named %s" % lockName)
        return S_OK()

    def _openAll(self):
        """
        WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
        DO NOT USE EXCEPT IN JUST SPAWNED NEW CHILD PROCESSES!!!!!!!!
        NEVER IN THE PARENT PROCESS!!!!!!
        WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
        """
        for lockName in self.__locks.keys():
            try:
                self.__locks[lockName].release()
            except (RuntimeError, thread.error, KeyError):
                pass

    def _setAllEvents(self):
        """
        WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
        DO NOT USE EXCEPT IN JUST SPAWNED NEW CHILD PROCESSES!!!!!!!!
        NEVER IN THE PARENT PROCESS!!!!!!
        WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
        """
        for evName in self.__events.keys():
            try:
                self.__events[evName].set()
            except KeyError:
                pass


if __name__ == "__main__":
    lr = LockRing()
    lock = lr.getLock("test1")
    print("ACQUIRING LOCK", lock)
    lock.acquire()
    print("IS THE SAME LOCK? ", lock == lr.getLock("test1"))
    print("OPENING ALL LOCKS")
    lr._openAll()
    print("REACQUIRING LOCK", lock)
    lr.acquire("test1")
    print("RELEASING LOCK")
    lr.release("test1")
    print("IS SINGLETON", lr == LockRing())
    ev = lr.getEvent("POT")
    ev.set()
    lr._setAllEvents()
    print("ALL OK")



""" BaseAction

  Base class for Actions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger

__RCSID__ = "$Id$"


class BaseAction(object):
    """
    Base class for all actions. It defines a constructor an a run main method.
    """

    def __init__(self, name, decisionParams, enforcementResult, singlePolicyResults, clients):

        # enforcementResult supposed to look like:
        # {
        #   'Status'        : <str>,
        #   'Reason'        : <str>,
        #   'PolicyActions' : <list>,
        #   [ 'EndDate' : <str> ]
        # }

        # decisionParams supposed to look like:
        # {
        #   'element'     : None,
        #   'name'        : None,
        #   'elementType' : None,
        #   'statusType'  : None,
        #   'status'      : None,
        #   'reason'      : None,
        #   'tokenOwner'  : None
        # }

        self.actionName = name  # 'BaseAction'
        self.decisionParams = decisionParams
        self.enforcementResult = enforcementResult
        self.singlePolicyResults = singlePolicyResults
        self.clients = clients
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.log.verbose("Running %s action" % self.__class__.__name__)

    def run(self):
        """
        Method to be over written by the real actions
        """

        self.log.warn("%s: you may want to overwrite this method" % self.actionName)



"""
Encoding and decoding for dirac, Ids:
 i -> int
 I -> long
 f -> float
 b -> bool
 s -> string
 z -> datetime
 n -> none
 l -> list
 t -> tuple
 d -> dictionary
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from past.builtins import long
import six
import datetime
import os

import functools
import inspect
import traceback

from collections import defaultdict
from pprint import pprint


def _ord(char):
    """Convert a single character string to it's byte value

    In Python 2 a single byte is represented as a string whereas in Python 3
    it is an integer. This function converts it as appropriate.
    """
    if six.PY2:
        return char
    else:
        return ord(char)


# This is a hack for Python 3 to make it possible to import DEncode
# There is not point in porting DEncode to Python 3 as it will be removed as
# part of the HTTPS transition.
class types(object):
    IntType = int
    LongType = long if six.PY2 else int
    FloatType = float
    BooleanType = bool
    StringType = str
    UnicodeType = type(u"")
    NoneType = type(None)
    ListType = list
    TupleType = tuple
    DictType = dict


# Setting this environment variable to any value will enable the dump of the debugging
# call stack
DIRAC_DEBUG_DENCODE_CALLSTACK = bool(os.environ.get("DIRAC_DEBUG_DENCODE_CALLSTACK", False))

# This global dictionary contains
# {<method name> : set (<class names)}
# (a method name can be reused in other classes)
# DO NOT EDIT BY HAND, use ignoreEncodeWarning decorator
DENCODE_WARNING_IGNORED_METHODS = defaultdict(set)

# Depth of the stack to look for with inspect
CONTEXT_DEPTH = 100


def ignoreEncodeWarning(meth):
    """Decorator to put around method that should not anymore throw warnings

    :warning: do not use around functions

    :warning: for a class method, put it after the @classmethod decorator

    :param meth: decorated method
    """

    @functools.wraps(meth)
    def inner(*args, **kwargs):
        """Add the method and the class name to the DENCODE_WARNING_IGNORED_METHODS dict"""

        # The first parameter in args is "self"
        # Find out the class Name
        objInst = args[0]
        className = objInst.__class__.__name__
        if className == "type":  # This happens for class method
            className = objInst.__name__

        # if the decorated method is an exported method, just remove the 'export_' bit
        methName = meth.__name__.replace("export_", "")

        # Add the method name and the object name to the dictionary
        DENCODE_WARNING_IGNORED_METHODS[methName].add(className)
        return meth(*args, **kwargs)

    return inner


def printDebugCallstack(headerMessage):
    """Prints information about the current stack as well as the caller parameters.
    The purpose of this method is to track down all the places in DIRAC that might
    not survive the change to JSON encoding.

    Some methods are ignored:

    * all the AccountingDB method: https://github.com/DIRACGrid/DIRAC/issues/4319
    * all the method in DENCODE_WARNING_IGNORED_METHODS (see ignoreEncodeWarning)

    :param headerMessage: message to be displayed first
    :returns: None

    """

    def stripArgs(frame):
        """Keeps only the parameters and their values from a frame

        :param frame: frame object

        :returns: dict {param name: value}
        """
        # Get all the arguments of the call
        allArgs = inspect.getargvalues(frame)
        # Keep only the arguments that are parameters of the call, as well as their value
        return dict([(argName, allArgs.locals[argName]) for argName in allArgs.args])

    tb = traceback.format_stack()
    frames = inspect.stack(context=CONTEXT_DEPTH)

    # Flag set to true only if we figure it's an RPC call
    # In that case, we display more info
    isRPCCall = False

    # For each entry in the stack, check if the method name is in the list of method to be ignored
    for frameRecord in reversed(frames):
        frameFuncName = frameRecord[3]
        # If the method is in the list of ignored method,
        # check that the method is from the good class
        if frameFuncName in DENCODE_WARNING_IGNORED_METHODS:
            try:
                # Take the frame object https://docs.python.org/2.7/reference/datamodel.html
                frameObj = frameRecord[0]
                # Check that the self attribute of the function points to a class which is listed
                # as to be ignored
                className = frameObj.f_locals["self"].__class__.__name__
                # if that is the case, then we return
                if className in DENCODE_WARNING_IGNORED_METHODS[frameFuncName]:
                    return
            # Exception may be thrown when trying to get the className
            except (KeyError, AttributeError):
                pass

        # Else, if we are answering an RPC call
        elif frameFuncName == "_executeAction":
            # This requires special handling because the only way to know
            # which method was called server side is to check at the proposalTuple

            frameObj = frameRecord[0]

            # The _executeAction method takes as parameter the handlerObj and the proposalTuple

            # Extract the method name from the proposalTuple
            funcName = frameObj.f_locals["proposalTuple"][1][1]

            # Extract the class name from the handlerObj
            className = frameObj.f_locals["handlerObj"].__class__.__name__

            if funcName in DENCODE_WARNING_IGNORED_METHODS and className in DENCODE_WARNING_IGNORED_METHODS[funcName]:
                return
            else:
                # If it is not to be ignored, save the parameters to display them
                isRPCCall = True
                rpcDetails = "RPC call service %s method %s" % (className, funcName)
                break

    # The datetime are encoded as tuple. Since datetime are taken care of
    # in JSerializer, just don't print a warning here
    # Note: -3 because we have to go past (de/encodeTuple and the Traceback module)
    if "encodeDateTime" in tb[-3] or "decodeDateTime" in tb[-3]:
        return

    # The accountingDB stores a encoding of the bucketsLength
    # this is ok for now, so silent all the AccountingDB error
    if any(["AccountingDB" in tr for tr in reversed(tb)]):
        return

    print("=" * 45, headerMessage, "=" * 45)
    # print the traceback that leads us here
    # remove the last element which is the traceback module call
    for line in tb[:-1]:
        print(line)

    # Now we try to navigate up to the caller of dEncode.
    # For this, we find the frame in which we enter dEncode.
    # We keep the parameters to display it.
    # Then we navigate to the parent frame, and we display the file
    # and line number where this call was done
    try:
        framesIter = iter(frames)
        for frame in framesIter:
            # First check that we are using either 'encode' or 'decode' function
            if frame[3] in ("encode", "decode"):
                # Then check it is the good file
                if frame[1].endswith("DIRAC/Core/Utilities/DEncode.py"):
                    # Keep the arguments of the DEncode call
                    dencArgs = stripArgs(frame[0])
                    # Take the calling frame
                    frame = next(framesIter)
                    print("Calling frame: %s" % (frame[1:3],))
                    if isRPCCall:
                        print(rpcDetails)
                    print("With arguments ", end=" ")
                    pprint(dencArgs)
                    break
    except Exception:
        pass
    print("=" * 100)
    print()
    print()


_dateTimeObject = datetime.datetime.utcnow()
_dateTimeType = type(_dateTimeObject)
_dateType = type(_dateTimeObject.date())
_timeType = type(_dateTimeObject.time())

g_dEncodeFunctions = {}
g_dDecodeFunctions = {}


def encodeInt(iValue, eList):
    """Encoding ints"""

    eList.extend((b"i", str(iValue).encode(), b"e"))


def decodeInt(data, i):
    """Decoding ints"""

    i += 1
    end = data.index(b"e", i)
    value = int(data[i:end])
    return (value, end + 1)


g_dEncodeFunctions[types.IntType] = encodeInt
g_dDecodeFunctions[_ord("i")] = decodeInt


def encodeLong(iValue, eList):
    """Encoding longs"""

    # corrected by KGG   eList.extend( ( "l", str( iValue ), "e" ) )
    eList.extend((b"I", str(iValue).encode(), b"e"))


def decodeLong(data, i):
    """Decoding longs"""

    i += 1
    end = data.index(_ord("e"), i)
    value = long(data[i:end])
    return (value, end + 1)


if not six.PY3:
    g_dEncodeFunctions[types.LongType] = encodeLong
g_dDecodeFunctions[_ord("I")] = decodeLong


def encodeFloat(iValue, eList):
    """Encoding floats"""

    eList.extend((b"f", str(iValue).encode(), b"e"))


def decodeFloat(data, i):
    """Decoding floats"""

    i += 1
    end = data.index(b"e", i)
    if end + 1 < len(data) and data[end + 1] in (_ord("+"), _ord("-")):
        eI = end
        end = data.index(b"e", end + 1)
        value = float(data[i:eI].decode()) * 10 ** int(data[eI + 1 : end].decode())
    else:
        value = float(data[i:end].decode())
    return (value, end + 1)


g_dEncodeFunctions[types.FloatType] = encodeFloat
g_dDecodeFunctions[_ord("f")] = decodeFloat


def encodeBool(bValue, eList):
    """Encoding booleans"""

    if bValue:
        eList.append(b"b1")
    else:
        eList.append(b"b0")


def decodeBool(data, i):
    """Decoding booleans"""

    if data[i + 1] == _ord("0"):
        return (False, i + 2)
    else:
        return (True, i + 2)


g_dEncodeFunctions[types.BooleanType] = encodeBool
g_dDecodeFunctions[_ord("b")] = decodeBool


def encodeString(sValue, eList):
    """Encoding strings"""
    if six.PY3 and not isinstance(sValue, bytes):
        sValue = sValue.encode()
    eList.extend((b"s", str(len(sValue)).encode(), b":", sValue))


def decodeString(data, i):
    """Decoding strings"""
    i += 1
    colon = data.index(b":", i)
    value = int(data[i:colon].decode())
    colon += 1
    end = colon + value
    retVal = data[colon:end]
    if six.PY3:
        retVal = retVal.decode(errors="surrogateescape")
    return (retVal, end)


g_dEncodeFunctions[types.StringType] = encodeString
g_dEncodeFunctions[bytes] = encodeString
g_dDecodeFunctions[_ord("s")] = decodeString


def encodeUnicode(sValue, eList):
    """Encoding unicode strings"""
    valueStr = sValue.encode("utf-8")
    eList.extend((b"u", str(len(valueStr)).encode(), b":", valueStr))


def decodeUnicode(data, i):
    """Decoding unicode strings"""

    i += 1
    colon = data.index(b":", i)
    value = int(data[i:colon])
    colon += 1
    end = colon + value
    return (six.text_type(data[colon:end].decode("utf-8")), end)


if six.PY2:
    g_dEncodeFunctions[types.UnicodeType] = encodeUnicode
    g_dDecodeFunctions[_ord("u")] = decodeUnicode
else:
    g_dDecodeFunctions[_ord("u")] = decodeString


def encodeDateTime(oValue, eList):
    """Encoding datetime"""

    if isinstance(oValue, _dateTimeType):
        tDateTime = (
            oValue.year,
            oValue.month,
            oValue.day,
            oValue.hour,
            oValue.minute,
            oValue.second,
            oValue.microsecond,
            oValue.tzinfo,
        )
        eList.append(b"za")
        # corrected by KGG encode( tDateTime, eList )
        g_dEncodeFunctions[type(tDateTime)](tDateTime, eList)
    elif isinstance(oValue, _dateType):
        tData = (oValue.year, oValue.month, oValue.day)
        eList.append(b"zd")
        # corrected by KGG encode( tData, eList )
        g_dEncodeFunctions[type(tData)](tData, eList)
    elif isinstance(oValue, _timeType):
        tTime = (oValue.hour, oValue.minute, oValue.second, oValue.microsecond, oValue.tzinfo)
        eList.append(b"zt")
        # corrected by KGG encode( tTime, eList )
        g_dEncodeFunctions[type(tTime)](tTime, eList)
    else:
        raise Exception("Unexpected type %s while encoding a datetime object" % str(type(oValue)))


def decodeDateTime(data, i):
    """Decoding datetime"""

    i += 1
    dataType = data[i]
    # corrected by KGG tupleObject, i = decode( data, i + 1 )
    tupleObject, i = g_dDecodeFunctions[data[i + 1]](data, i + 1)
    if dataType == _ord("a"):
        dtObject = datetime.datetime(*tupleObject)
    elif dataType == _ord("d"):
        dtObject = datetime.date(*tupleObject)
    elif dataType == _ord("t"):
        dtObject = datetime.time(*tupleObject)
    else:
        raise Exception("Unexpected type %s while decoding a datetime object" % dataType)
    return (dtObject, i)


g_dEncodeFunctions[_dateTimeType] = encodeDateTime
g_dEncodeFunctions[_dateType] = encodeDateTime
g_dEncodeFunctions[_timeType] = encodeDateTime
g_dDecodeFunctions[_ord("z")] = decodeDateTime


def encodeNone(_oValue, eList):
    """Encoding None"""

    eList.append(b"n")


def decodeNone(_data, i):
    """Decoding None"""

    return (None, i + 1)


g_dEncodeFunctions[types.NoneType] = encodeNone
g_dDecodeFunctions[_ord("n")] = decodeNone


def encodeList(lValue, eList):
    """Encoding list"""

    eList.append(b"l")
    for uObject in lValue:
        g_dEncodeFunctions[type(uObject)](uObject, eList)
    eList.append(b"e")


def decodeList(data, i):
    """Decoding list"""

    oL = []
    i += 1
    while data[i] != _ord("e"):
        ob, i = g_dDecodeFunctions[data[i]](data, i)
        oL.append(ob)
    return (oL, i + 1)


g_dEncodeFunctions[types.ListType] = encodeList
g_dDecodeFunctions[_ord("l")] = decodeList


def encodeTuple(lValue, eList):
    """Encoding tuple"""

    if DIRAC_DEBUG_DENCODE_CALLSTACK:
        printDebugCallstack("Encoding tuples")

    eList.append(b"t")
    for uObject in lValue:
        g_dEncodeFunctions[type(uObject)](uObject, eList)
    eList.append(b"e")


def decodeTuple(data, i):
    """Decoding tuple"""

    if DIRAC_DEBUG_DENCODE_CALLSTACK:
        printDebugCallstack("Decoding tuples")

    oL, i = decodeList(data, i)
    return (tuple(oL), i)


g_dEncodeFunctions[types.TupleType] = encodeTuple
g_dDecodeFunctions[_ord("t")] = decodeTuple


def encodeDict(dValue, eList):
    """Encoding dictionary"""

    if DIRAC_DEBUG_DENCODE_CALLSTACK:
        # If we have numbers as keys
        if any([isinstance(x, six.integer_types + (float,)) for x in dValue]):
            printDebugCallstack("Encoding dict with numeric keys")

    eList.append(b"d")
    for key in sorted(dValue):
        g_dEncodeFunctions[type(key)](key, eList)
        g_dEncodeFunctions[type(dValue[key])](dValue[key], eList)
    eList.append(b"e")


def decodeDict(data, i):
    """Decoding dictionary"""

    oD = {}
    i += 1
    while data[i] != _ord("e"):

        if DIRAC_DEBUG_DENCODE_CALLSTACK:
            # If we have numbers as keys
            if data[i] in (_ord("i"), _ord("I"), _ord("f")):
                printDebugCallstack("Decoding dict with numeric keys")

        k, i = g_dDecodeFunctions[data[i]](data, i)
        oD[k], i = g_dDecodeFunctions[data[i]](data, i)
    return (oD, i + 1)


g_dEncodeFunctions[types.DictType] = encodeDict
g_dDecodeFunctions[_ord("d")] = decodeDict


# Encode function
def encode(uObject):
    """Generic encoding function"""
    eList = []
    # print("ENCODE FUNCTION : %s" % g_dEncodeFunctions[ type( uObject ) ])
    g_dEncodeFunctions[type(uObject)](uObject, eList)
    return b"".join(eList)


def decode(data):
    """Generic decoding function"""
    if not data:
        return data
    # print("DECODE FUNCTION : %s" % g_dDecodeFunctions[ sStream [ iIndex ] ])
    if not isinstance(data, bytes):
        raise NotImplementedError("This should never happen")
    return g_dDecodeFunctions[data[0]](data, 0)


if __name__ == "__main__":
    gObject = {2: "3", True: (3, None), 2.0 * 10 ** 20: 2.0 * 10 ** -10}
    print("Initial: %s" % gObject)
    gData = encode(gObject)
    print("Encoded: %s" % gData)
    print("Decoded: %s, [%s]" % decode(gData))



""" Utilities to process parametric job definitions and generate
    bunches of parametric jobs. It exposes the following functions:

    getParameterVectorLength() - to get the total size of the bunch of parametric jobs
    generateParametricJobs() - to get a list of expanded descriptions of all the jobs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import re

from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities.DErrno import EWMSJDL


def __getParameterSequence(nPar, parList=[], parStart=1, parStep=0, parFactor=1):

    if parList:
        if nPar != len(parList):
            return []
        else:
            parameterList = list(parList)
    else:
        # The first parameter must have the same type as the other ones even if not defined explicitly
        parameterList = [parStart * type(parFactor)(1) + type(parStep)(0)]
        for np in range(1, nPar):
            parameterList.append(parameterList[np - 1] * parFactor + parStep)

    return parameterList


def getParameterVectorLength(jobClassAd):
    """Get the length of parameter vector in the parametric job description

    :param jobClassAd: ClassAd job description object
    :return: result structure with the Value: int number of parameter values, None if not a parametric job
    """

    nParValues = None
    attributes = jobClassAd.getAttributes()
    for attribute in attributes:
        if attribute.startswith("Parameters"):
            if jobClassAd.isAttributeList(attribute):
                parameterList = jobClassAd.getListFromExpression(attribute)
                nThisParValues = len(parameterList)
            else:
                nThisParValues = jobClassAd.getAttributeInt(attribute)
            if nParValues is not None and nParValues != nThisParValues:
                return S_ERROR(
                    EWMSJDL,
                    "Different length of parameter vectors: for %s, %s != %d" % (attribute, nParValues, nThisParValues),
                )
            nParValues = nThisParValues
    if nParValues is not None and nParValues <= 0:
        return S_ERROR(EWMSJDL, "Illegal number of job parameters %d" % (nParValues))
    return S_OK(nParValues)


def __updateAttribute(classAd, attribute, parName, parValue):

    # If there is something to do:
    pattern = r"%%\(%s\)s" % parName
    if parName == "0":
        pattern = "%s"
    expr = classAd.get_expression(attribute)
    if not re.search(pattern, expr):
        return False

    pattern = "%%(%s)s" % parName
    if parName == "0":
        pattern = "%s"

    parValue = parValue.strip()
    if classAd.isAttributeList(attribute):
        parValue = parValue.strip()
        if parValue.startswith("{"):
            parValue = parValue.lstrip("{").rstrip("}").strip()

    expr = classAd.get_expression(attribute)
    newexpr = expr.replace(pattern, str(parValue))
    classAd.set_expression(attribute, newexpr)
    return True


def generateParametricJobs(jobClassAd):
    """Generate a series of ClassAd job descriptions expanding
        job parameters

    :param jobClassAd: ClassAd job description object
    :return: list of ClassAd job description objects
    """
    if not jobClassAd.lookupAttribute("Parameters"):
        return S_OK([jobClassAd.asJDL()])

    result = getParameterVectorLength(jobClassAd)
    if not result["OK"]:
        return result
    nParValues = result["Value"]
    if nParValues is None:
        return S_ERROR(EWMSJDL, "Can not determine the number of job parameters")

    parameterDict = {}
    attributes = jobClassAd.getAttributes()
    for attribute in attributes:
        for key in ["Parameters", "ParameterStart", "ParameterStep", "ParameterFactor"]:
            if attribute.startswith(key):
                seqID = "0" if "." not in attribute else attribute.split(".")[1]
                parameterDict.setdefault(seqID, {})
                if key == "Parameters":
                    if jobClassAd.isAttributeList(attribute):
                        parList = jobClassAd.getListFromExpression(attribute)
                        if len(parList) != nParValues:
                            return S_ERROR(EWMSJDL, "Inconsistent parametric job description")
                        parameterDict[seqID]["ParameterList"] = parList
                    else:
                        if attribute != "Parameters":
                            return S_ERROR(EWMSJDL, "Inconsistent parametric job description")
                        nPar = jobClassAd.getAttributeInt(attribute)
                        if nPar is None:
                            value = jobClassAd.get_expression(attribute)
                            return S_ERROR(
                                EWMSJDL, "Inconsistent parametric job description: %s=%s" % (attribute, value)
                            )
                        parameterDict[seqID]["Parameters"] = nPar
                else:
                    value = jobClassAd.getAttributeInt(attribute)
                    if value is None:
                        value = jobClassAd.getAttributeFloat(attribute)
                        if value is None:
                            value = jobClassAd.get_expression(attribute)
                            return S_ERROR("Illegal value for %s JDL field: %s" % (attribute, value))
                    parameterDict[seqID][key] = value

    if "0" in parameterDict and not parameterDict.get("0"):
        parameterDict.pop("0")

    parameterLists = {}
    for seqID in parameterDict:
        parList = __getParameterSequence(
            nParValues,
            parList=parameterDict[seqID].get("ParameterList", []),
            parStart=parameterDict[seqID].get("ParameterStart", 1),
            parStep=parameterDict[seqID].get("ParameterStep", 0),
            parFactor=parameterDict[seqID].get("ParameterFactor", 1),
        )
        if not parList:
            return S_ERROR(EWMSJDL, "Inconsistent parametric job description")

        parameterLists[seqID] = parList

    jobDescList = []
    jobDesc = jobClassAd.asJDL()
    # Width of the sequential parameter number
    zLength = len(str(nParValues - 1))
    for n in range(nParValues):
        newJobDesc = jobDesc
        newJobDesc = newJobDesc.replace("%n", str(n).zfill(zLength))
        newClassAd = ClassAd(newJobDesc)
        for seqID in parameterLists:
            parameter = parameterLists[seqID][n]
            for attribute in newClassAd.getAttributes():
                __updateAttribute(newClassAd, attribute, seqID, str(parameter))

        for seqID in parameterLists:
            for attribute in ["Parameters", "ParameterStart", "ParameterStep", "ParameterFactor"]:
                if seqID == "0":
                    newClassAd.deleteAttribute(attribute)
                else:
                    newClassAd.deleteAttribute("%s.%s" % (attribute, seqID))

            parameter = parameterLists[seqID][n]
            if seqID == "0":
                attribute = "Parameter"
            else:
                attribute = "Parameter.%s" % seqID
            if isinstance(parameter, six.string_types) and parameter.startswith("{"):
                newClassAd.insertAttributeInt(attribute, str(parameter))
            else:
                newClassAd.insertAttributeString(attribute, str(parameter))

        newClassAd.insertAttributeInt("ParameterNumber", n)
        newJDL = newClassAd.asJDL()
        jobDescList.append(newJDL)

    return S_OK(jobDescList)



########################################################################
# Author : Andrei Tsaregorodtsev
########################################################################
"""
  Utilities for managing DIRAC configuration:
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from copy import deepcopy
import socket
import six

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOs, getVOOption
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getDIRACSiteName
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.Core.Utilities.Glue2 import getGlue2CEInfo
from DIRAC.Core.Utilities.SiteSEMapping import getSEHosts
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers


def getGridVOs():
    """Get all the VOMS VO names served by this DIRAC service

    :return: S_OK(list)/S_ERROR()
    """
    voNames = []
    result = getVOs()
    if not result["OK"]:
        return result
    else:
        vos = result["Value"]
        for vo in vos:
            vomsVO = getVOOption(vo, "VOMSName")
            if vomsVO:
                voNames.append(vomsVO)
    return S_OK(voNames)


def getGridCEs(vo, bdiiInfo=None, ceBlackList=None, hostURL=None):
    """Get all the CEs available for a given VO and having queues in Production state

    :param str vo: VO name
    :param dict bddiInfo: information from BDII
    :param list ceBlackList: CEs from black list
    :param str hostURL: host URL

    :return: Dictionary with keys: OK, Value, BdiiInfo, UnknownCEs
    """
    knownCEs = set()
    cesInInformation = set()
    if ceBlackList is not None:
        knownCEs = knownCEs.union(set(ceBlackList))

    ceBdiiDict = bdiiInfo
    if bdiiInfo is None:
        result = getGlue2CEInfo(vo, host=hostURL)
        if not result["OK"]:
            return result
        ceBdiiDict = result["Value"]

    siteDict = {}
    for site in ceBdiiDict:
        siteCEs = set(ceBdiiDict[site]["CEs"].keys())
        cesInInformation.update(siteCEs)
        newCEs = siteCEs - knownCEs
        if not newCEs:
            continue

        ceFullDict = {}
        for ce in newCEs:
            ceDict = {}
            ceInfo = ceBdiiDict[site]["CEs"][ce]
            ceType = "Unknown"
            ceDict["Queues"] = []
            for queue in ceInfo["Queues"]:
                queueStatus = ceInfo["Queues"][queue].get("GlueCEStateStatus", "UnknownStatus")
                if "production" in queueStatus.lower():
                    ceType = ceInfo["Queues"][queue].get("GlueCEImplementationName", "")
                    ceDict["Queues"].append(queue)
            if not ceDict["Queues"]:
                continue

            ceDict["CEType"] = ceType
            ceDict["GOCSite"] = site
            ceDict["CEID"] = ce
            systemName = ceInfo.get("GlueHostOperatingSystemName", "Unknown")
            systemVersion = ceInfo.get("GlueHostOperatingSystemVersion", "Unknown")
            systemRelease = ceInfo.get("GlueHostOperatingSystemRelease", "Unknown")
            ceDict["System"] = (systemName, systemVersion, systemRelease)

            ceFullDict[ce] = ceDict

        siteDict[site] = ceFullDict

    result = S_OK(siteDict)
    result["BdiiInfo"] = ceBdiiDict

    unknownCEs = knownCEs - cesInInformation
    result["UnknownCEs"] = unknownCEs
    return result


def getSiteUpdates(vo, bdiiInfo=None, log=None, onecore=False):
    """Get all the necessary updates for the already defined sites and CEs

    :param str vo: VO name
    :param dict bdiiInfo: information from DBII
    :param object log: logger
    :param bool onecore: whether to add single core copies of multicore queues, see the documentation about :ref:`CE`
       and the :mod:`~DIRAC.ConfigurationSystem.Agent.Bdii2CSAgent` configuration for details

    :result: S_OK(set)/S_ERROR()
    """

    def addToChangeSet(entry, changeSet):
        """Inner function to update changeSet with entry (a tuple)

        :param tuple entry: entry to add to changeSet
        :param set changeSet: set collecting stuff to change
        """
        _section, _option, value, new_value = entry
        if new_value and new_value != value:
            changeSet.add(entry)

    def dropTag(tags, tagToDrop):
        """Remove tag from a comma-separated string of tags.

        :param str tags: the string of current tags
        :param str tagToDrop: the tag to potentially remove
        :return: string of comma separated tags
        """
        return ",".join(sorted(set(tags.split(",")).difference({tagToDrop}))).strip(",")

    def addTag(tags, tagToAdd):
        """Add tag to a comma-separated string of tags.

        :param str tags: the string of current tags
        :param str tagToAdd: the tag to potentially add
        :return: string of comma separated tags
        """
        return ",".join(sorted(set(tags.split(",")).union({tagToAdd}))).strip(",")

    if log is None:
        log = gLogger

    ceBdiiDict = bdiiInfo
    if bdiiInfo is None:
        result = getGlue2CEInfo(vo)
        if not result["OK"]:
            return result
        ceBdiiDict = result["Value"]

    if onecore:
        # If enabled this creates a copy of the queue with multiple processors and sets NumberOfProcessors to 1 for ARC
        # and HTCondorCE entries

        def makeNewQueueName(queueName, ceType):
            """Create a new queueName for single core queues."""
            if ceType == "HTCondorCE":
                return queueName + "1core"
            # we should have only ARC left, we add 1core to the middle part
            queueNameSplit = queueName.split("-", 2)
            queueNameSplit[1] = queueNameSplit[1] + "1core"
            return "-".join(queueNameSplit)

        for siteName, ceDict in ceBdiiDict.items():
            for _ceName, ceInfo in ceDict["CEs"].items():
                newQueues = dict()
                for queueName, queueDict in ceInfo["Queues"].items():
                    if (
                        queueDict["GlueCEImplementationName"] not in ("ARC", "HTCondorCE")
                        or int(queueDict.get("NumberOfProcessors", 1)) == 1
                    ):
                        continue
                    newQueueName = makeNewQueueName(queueName, queueDict["GlueCEImplementationName"])
                    newQueueDict = deepcopy(queueDict)
                    newQueueDict["NumberOfProcessors"] = 1
                    newQueues[newQueueName] = newQueueDict

                ceInfo["Queues"].update(newQueues)

    changeSet = set()
    for site in ceBdiiDict:
        result = getDIRACSiteName(site)
        if not result["OK"]:
            continue
        siteNames = result["Value"]
        for siteName in siteNames:
            siteSection = cfgPath("/Resources", "Sites", siteName.split(".")[0], siteName)
            result = gConfig.getOptionsDict(siteSection)
            if not result["OK"]:
                continue
            siteDict = result["Value"]
            # Current CS values
            coor = siteDict.get("Coordinates", "Unknown")
            mail = siteDict.get("Mail", "Unknown").replace(" ", "")
            description = siteDict.get("Description", "Unknown")
            description = description.replace(" ,", ",")

            longitude = ceBdiiDict[site].get("GlueSiteLongitude", "").strip()
            latitude = ceBdiiDict[site].get("GlueSiteLatitude", "").strip()

            # Current BDII value
            newcoor = ""
            if longitude and latitude:
                newcoor = "%s:%s" % (longitude, latitude)
            newmail = ceBdiiDict[site].get("GlueSiteSysAdminContact", "").replace("mailto:", "").strip()
            newdescription = ceBdiiDict[site].get("GlueSiteDescription", "").strip()
            newdescription = ", ".join([line.strip() for line in newdescription.split(",")])

            # Adding site data to the changes list
            addToChangeSet((siteSection, "Coordinates", coor, newcoor), changeSet)
            addToChangeSet((siteSection, "Mail", mail, newmail), changeSet)
            addToChangeSet((siteSection, "Description", description, newdescription), changeSet)

            ces = gConfig.getSections(cfgPath(siteSection, "CEs"))
            for ce in ces.get("Value", []):
                ceSection = cfgPath(siteSection, "CEs", ce)
                ceDict = {}
                result = gConfig.getOptionsDict(ceSection)
                if result["OK"]:
                    ceDict = result["Value"]
                else:
                    if ceBdiiDict[site]["CEs"].get(ce, None):
                        log.notice("Adding new CE", "%s to site %s/%s" % (ce, siteName, site))
                ceInfo = ceBdiiDict[site]["CEs"].get(ce, None)
                if ceInfo is None:
                    ceType = ceDict.get("CEType", "")
                    continue

                # Current CS CE info
                arch = ceDict.get("architecture", "Unknown")
                OS = ceDict.get("OS", "Unknown")
                si00 = ceDict.get("SI00", "Unknown")
                ceType = ceDict.get("CEType", "Unknown")
                ram = ceDict.get("MaxRAM", "Unknown")
                submissionMode = ceDict.get("SubmissionMode", "Unknown")

                # Current BDII CE info
                newarch = ceBdiiDict[site]["CEs"][ce].get("GlueHostArchitecturePlatformType", "").strip()
                systemName = ceInfo.get("GlueHostOperatingSystemName", "").strip()
                systemVersion = ceInfo.get("GlueHostOperatingSystemVersion", "").strip()
                systemRelease = ceInfo.get("GlueHostOperatingSystemRelease", "").strip()
                newOS = ""
                if systemName and systemVersion and systemRelease:
                    newOS = "_".join((systemName, systemVersion, systemRelease))
                newsi00 = ceInfo.get("GlueHostBenchmarkSI00", "").strip()
                newCEType = "Unknown"
                for queue in ceInfo["Queues"]:
                    queueDict = ceInfo["Queues"][queue]
                    newCEType = queueDict.get("GlueCEImplementationName", "").strip()
                    if newCEType:
                        break
                if newCEType == "ARC-CE":
                    newCEType = "ARC"

                newSubmissionMode = None
                if newCEType in ["ARC", "CREAM"]:
                    newSubmissionMode = "Direct"
                newRAM = ceInfo.get("GlueHostMainMemoryRAMSize", "").strip()
                # Protect from unreasonable values
                if newRAM and int(newRAM) > 150000:
                    newRAM = ""

                # Adding CE data to the change list
                addToChangeSet((ceSection, "architecture", arch, newarch), changeSet)
                addToChangeSet((ceSection, "OS", OS, newOS), changeSet)
                addToChangeSet((ceSection, "SI00", si00, newsi00), changeSet)
                addToChangeSet((ceSection, "CEType", ceType, newCEType), changeSet)
                addToChangeSet((ceSection, "MaxRAM", ram, newRAM), changeSet)
                if submissionMode == "Unknown" and newSubmissionMode:
                    addToChangeSet((ceSection, "SubmissionMode", submissionMode, newSubmissionMode), changeSet)

                for queue, queueInfo in ceInfo["Queues"].items():
                    queueStatus = queueInfo["GlueCEStateStatus"]
                    queueSection = cfgPath(ceSection, "Queues", queue)
                    queueDict = {}
                    result = gConfig.getOptionsDict(queueSection)
                    if result["OK"]:
                        queueDict = result["Value"]
                    else:
                        if queueStatus.lower() == "production":
                            log.notice("Adding new queue", "%s to CE %s" % (queue, ce))
                        else:
                            continue

                    # Current CS queue info
                    maxCPUTime = queueDict.get("maxCPUTime", "Unknown")
                    si00 = queueDict.get("SI00", "Unknown")
                    maxTotalJobs = queueDict.get("MaxTotalJobs", "Unknown")

                    # Current BDII queue info
                    newMaxCPUTime = queueInfo.get("GlueCEPolicyMaxCPUTime", "")
                    if newMaxCPUTime == "4" * len(newMaxCPUTime) or newMaxCPUTime == "9" * len(newMaxCPUTime):
                        newMaxCPUTime = ""
                    wallTime = queueInfo.get("GlueCEPolicyMaxWallClockTime", "")
                    if wallTime == "4" * len(wallTime) or wallTime == "9" * len(wallTime):
                        wallTime = ""
                    if wallTime and int(wallTime) > 0:
                        if not newMaxCPUTime:
                            newMaxCPUTime = str(int(0.8 * int(wallTime)))
                        else:
                            if int(wallTime) <= int(newMaxCPUTime):
                                newMaxCPUTime = str(int(0.8 * int(wallTime)))
                    newSI00 = ""
                    caps = queueInfo.get("GlueCECapability", [])
                    if isinstance(caps, six.string_types):
                        caps = [caps]
                    for cap in caps:
                        if "CPUScalingReferenceSI00" in cap:
                            newSI00 = cap.split("=")[-1]

                    # tags, processors, localCEType
                    tag = queueDict.get("Tag", "")
                    reqTag = queueDict.get("RequiredTag", "")
                    # LocalCEType can be empty (equivalent to "InProcess")
                    # or "Pool", "Singularity", but also "Pool/Singularity"
                    localCEType = queueDict.get("LocalCEType", "")
                    try:
                        localCEType_inner = localCEType.split("/")[1]
                    except IndexError:
                        localCEType_inner = ""

                    numberOfProcessors = int(queueDict.get("NumberOfProcessors", 0))
                    newNOP = int(queueInfo.get("NumberOfProcessors", 1))

                    # Adding queue info to the CS
                    addToChangeSet((queueSection, "maxCPUTime", maxCPUTime, newMaxCPUTime), changeSet)
                    addToChangeSet((queueSection, "SI00", si00, newSI00), changeSet)

                    # add RequiredTag if onecore is enabled, do this here for previously created MultiCore queues
                    if newNOP > 1 and onecore:
                        addToChangeSet(
                            (queueSection, "RequiredTag", reqTag, addTag(reqTag, "MultiProcessor")), changeSet
                        )

                    if newNOP != numberOfProcessors:
                        addToChangeSet((queueSection, "NumberOfProcessors", numberOfProcessors, newNOP), changeSet)
                        if newNOP > 1:
                            # if larger than one, add MultiProcessor to site tags, and LocalCEType=Pool
                            addToChangeSet((queueSection, "Tag", tag, addTag(tag, "MultiProcessor")), changeSet)

                            if localCEType_inner:
                                newLocalCEType = "Pool/" + localCEType_inner
                            else:
                                newLocalCEType = "Pool"
                            addToChangeSet((queueSection, "LocalCEType", localCEType, newLocalCEType), changeSet)
                        else:
                            # if not larger than one, drop MultiProcessor Tag.
                            # Here we do not change the LocalCEType as Pool CE would still be perfectly valid.
                            changeSet.add((queueSection, "Tag", tag, dropTag(tag, "MultiProcessor")))
                            if onecore:
                                changeSet.add((queueSection, "RequiredTag", reqTag, dropTag(reqTag, "MultiProcessor")))

                    if maxTotalJobs == "Unknown":
                        newTotalJobs = min(1000, int(int(queueInfo.get("GlueCEInfoTotalCPUs", 0)) / 2))
                        newWaitingJobs = max(2, int(newTotalJobs * 0.1))
                        newTotalJobs = str(newTotalJobs)
                        newWaitingJobs = str(newWaitingJobs)
                        addToChangeSet((queueSection, "MaxTotalJobs", "", newTotalJobs), changeSet)
                        addToChangeSet((queueSection, "MaxWaitingJobs", "", newWaitingJobs), changeSet)

                    # Updating eligible VO list
                    VOs = set()
                    if queueDict.get("VO", ""):
                        VOs = set([q.strip() for q in queueDict.get("VO", "").split(",") if q])
                    if vo not in VOs:
                        VOs.add(vo)
                        VOs = list(VOs)
                        newVOs = ",".join(VOs)
                        addToChangeSet((queueSection, "VO", "", newVOs), changeSet)

    return S_OK(changeSet)


def getDBParameters(fullname):
    """Retrieve Database parameters from CS

    :param str fullname: should be of the form <System>/<DBname>
           defaultHost is the host to return if the option is not found in the CS.
           Not used as the method will fail if it cannot be found
           defaultPort is the port to return if the option is not found in the CS
           defaultUser is the user to return if the option is not found in the CS.
           Not usePassword is the password to return if the option is not found in the CS.
           Not used as the method will fail if it cannot be found
           defaultDB is the db to return if the option is not found in the CS.
           Not used as the method will fail if it cannot be found
           defaultQueueSize is the QueueSize to return if the option is not found in the CS

    :return: S_OK(dict)/S_ERROR() - dictionary with the keys: 'host', 'port', 'user', 'password',
                                    'db' and 'queueSize'
    """

    cs_path = getDatabaseSection(fullname)
    parameters = {}

    result = gConfig.getOption(cs_path + "/Host")
    if not result["OK"]:
        # No host name found, try at the common place
        result = gConfig.getOption("/Systems/Databases/Host")
        if not result["OK"]:
            return S_ERROR("Failed to get the configuration parameter: Host")
    dbHost = result["Value"]
    # Check if the host is the local one and then set it to 'localhost' to use
    # a socket connection
    if dbHost != "localhost":
        localHostName = socket.getfqdn()
        if localHostName == dbHost:
            dbHost = "localhost"
    parameters["Host"] = dbHost

    # Mysql standard
    dbPort = 3306
    result = gConfig.getOption(cs_path + "/Port")
    if not result["OK"]:
        # No individual port number found, try at the common place
        result = gConfig.getOption("/Systems/Databases/Port")
        if result["OK"]:
            dbPort = int(result["Value"])
    else:
        dbPort = int(result["Value"])
    parameters["Port"] = dbPort

    result = gConfig.getOption(cs_path + "/User")
    if not result["OK"]:
        # No individual user name found, try at the common place
        result = gConfig.getOption("/Systems/Databases/User")
        if not result["OK"]:
            return S_ERROR("Failed to get the configuration parameter: User")
    dbUser = result["Value"]
    parameters["User"] = dbUser

    result = gConfig.getOption(cs_path + "/Password")
    if not result["OK"]:
        # No individual password found, try at the common place
        result = gConfig.getOption("/Systems/Databases/Password")
        if not result["OK"]:
            return S_ERROR("Failed to get the configuration parameter: Password")
    dbPass = result["Value"]
    parameters["Password"] = dbPass

    result = gConfig.getOption(cs_path + "/DBName")
    if not result["OK"]:
        return S_ERROR("Failed to get the configuration parameter: DBName")
    dbName = result["Value"]
    parameters["DBName"] = dbName

    return S_OK(parameters)


def getElasticDBParameters(fullname):
    """Retrieve Database parameters from CS

    :param str fullname: should be of the form <System>/<DBname>

    :return: S_OK(dict)/S_ERROR()
    """

    cs_path = getDatabaseSection(fullname)
    parameters = {}

    result = gConfig.getOption(cs_path + "/Host")
    if not result["OK"]:
        # No host name found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/Host")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: Host. Using localhost")
            dbHost = "localhost"
        else:
            dbHost = result["Value"]
    else:
        dbHost = result["Value"]
    # Check if the host is the local one and then set it to 'localhost' to use
    # a socket connection
    if dbHost != "localhost":
        localHostName = socket.getfqdn()
        if localHostName == dbHost:
            dbHost = "localhost"
    parameters["Host"] = dbHost

    # Elasticsearch standard port
    result = gConfig.getOption(cs_path + "/Port")
    if not result["OK"]:
        # No individual port number found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/Port")
        if not result["OK"]:
            gLogger.warn("No configuration parameter set for Port, assuming URL points to right location")
            dbPort = None
        else:
            dbPort = int(result["Value"])
    else:
        dbPort = int(result["Value"])
    parameters["Port"] = dbPort

    result = gConfig.getOption(cs_path + "/User")
    if not result["OK"]:
        # No individual user name found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/User")
        if not result["OK"]:
            gLogger.warn(
                "Failed to get the configuration parameter: User. Assuming no user/password is provided/needed"
            )
            dbUser = None
        else:
            dbUser = result["Value"]
    else:
        dbUser = result["Value"]
    parameters["User"] = dbUser

    result = gConfig.getOption(cs_path + "/Password")
    if not result["OK"]:
        # No individual password found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/Password")
        if not result["OK"]:
            gLogger.warn(
                "Failed to get the configuration parameter: Password. Assuming no user/password is provided/needed"
            )
            dbPass = None
        else:
            dbPass = result["Value"]
    else:
        dbPass = result["Value"]
    parameters["Password"] = dbPass

    result = gConfig.getOption(cs_path + "/SSL")
    if not result["OK"]:
        # No SSL option found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/SSL")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: SSL. Assuming SSL is needed")
            ssl = True
        else:
            ssl = False if result["Value"].lower() in ("false", "no", "n") else True
    else:
        ssl = False if result["Value"].lower() in ("false", "no", "n") else True
    parameters["SSL"] = ssl

    # Elasticsearch use certs
    result = gConfig.getOption(cs_path + "/CRT")
    if not result["OK"]:
        # No CRT option found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/CRT")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: CRT. Using False")
            certs = False
        else:
            certs = result["Value"]
    else:
        certs = result["Value"]
    parameters["CRT"] = certs

    # Elasticsearch ca_certs
    result = gConfig.getOption(cs_path + "/ca_certs")
    if not result["OK"]:
        # No CA certificate found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/ca_certs")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: ca_certs. Using None")
            ca_certs = None
        else:
            ca_certs = result["Value"]
    else:
        ca_certs = result["Value"]
    parameters["ca_certs"] = ca_certs

    # Elasticsearch client_key
    result = gConfig.getOption(cs_path + "/client_key")
    if not result["OK"]:
        # No client private key found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/client_key")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: client_key. Using None")
            client_key = None
        else:
            client_key = result["Value"]
    else:
        client_key = result["Value"]
    parameters["client_key"] = client_key

    # Elasticsearch client_cert
    result = gConfig.getOption(cs_path + "/client_cert")
    if not result["OK"]:
        # No cient certificate found, try at the common place
        result = gConfig.getOption("/Systems/NoSQLDatabases/client_cert")
        if not result["OK"]:
            gLogger.warn("Failed to get the configuration parameter: client_cert. Using None")
            client_cert = None
        else:
            client_cert = result["Value"]
    else:
        client_cert = result["Value"]
    parameters["client_cert"] = client_cert

    return S_OK(parameters)


def getOAuthAPI(instance="Production"):
    """Get OAuth API url

    :param str instance: instance

    :return: str
    """
    return gConfig.getValue("/Systems/Framework/%s/URLs/OAuthAPI" % instance)


def getDIRACGOCDictionary():
    """
    Create a dictionary containing DIRAC site names and GOCDB site names
    using a configuration provided by CS.

    :return:  A dictionary of DIRAC site names (key) and GOCDB site names (value).
    """

    log = gLogger.getSubLogger("getDIRACGOCDictionary")
    log.debug("Begin function ...")

    result = gConfig.getConfigurationTree("/Resources/Sites", "Name")
    if not result["OK"]:
        log.error("getConfigurationTree() failed with message: %s" % result["Message"])
        return S_ERROR("Configuration is corrupted")
    siteNamesTree = result["Value"]

    dictionary = dict()
    PATHELEMENTS = 6  # site names have 6 elements in the path, i.e.:
    #    /Resource/Sites/<GRID NAME>/<DIRAC SITE NAME>/Name
    # [0]/[1]     /[2]  /[3]        /[4]              /[5]

    for path, gocdbSiteName in siteNamesTree.items():  # can be an iterator
        elements = path.split("/")
        if len(elements) != PATHELEMENTS:
            continue

        diracSiteName = elements[PATHELEMENTS - 2]
        dictionary[diracSiteName] = gocdbSiteName

    log.debug("End function.")
    return S_OK(dictionary)
